import abc
import logging
from typing import Optional, Union

from pyiron_contrib.tinybase.task import AbstractTask
from pyiron_contrib.tinybase.storage import (
    Storable,
    GenericStorage,
    pickle_load,
    pickle_dump,
)
from pyiron_contrib.tinybase.executor import (
    Executor,
    ExecutionContext,
    BackgroundExecutor,
    ProcessExecutor,
)
from pyiron_contrib.tinybase.database import DatabaseEntry
from pyiron_contrib.tinybase.project import ProjectInterface, ProjectAdapter
from pyiron_base.state import state


class TinyJob(Storable):
    """
    A tiny job unifies an executor, a task and its output.

    The job adds the task to the database and persists its input and output in a storage location.

    The input of the task is available from :attr:`~.input`. After the job has
    finished the output of the task can be
    accessed from :attr:`~.output` and the data written to storage from :attr:`.~storage`.

    This is an abstracat base class that works with any execution task without specifying it.  To create specialized
    jobs you can derive from it and overload :meth:`._get_task()` to return an
    instance of the task, e.g.

    >>> from somewhere import MyTask
    >>> class MyJob(TinyJob):
    ...     def _get_task(self):
    ...         return MyTask()

    The return value of :meth:`._get_task()` is persisted during the life time of the job.

    You can use :class:`.GenericTinyJob` to dynamically specify which task the job should execute.
    """

    def __init__(self, task: AbstractTask, project: ProjectInterface, job_name: str):
        """
        Create a new job.

        Args:
            task (:class:`.AbstractTask`): the underlying task to run
            project (:class:`.ProjectInterface`): the project the job should live in
            job_name (str): the name of the job.
        """
        if not isinstance(project, ProjectInterface):
            project = ProjectAdapter(project)
        self._project = project
        self._name = job_name
        self._task = task
        self._output = None
        self._storage = None
        self._executor = None
        self._id = None

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    def _update_id(self):
        self._id = self.project.get_job_id(self.name)

    @property
    def project(self):
        return self._project

    @property
    def status(self):
        try:
            return self._project.database.get_item(self.id).status
        except ValueError:
            return "initialized"

    @property
    def task(self):
        return self._task

    @property
    def jobtype(self):
        return self.task.__class__.__name__

    @property
    def input(self):
        return self.task.input

    @property
    def output(self):
        if self._output is not None:
            return self._output
        else:
            raise RuntimeError("Job not finished yet!")

    @property
    def storage(self):
        if self._storage is None:
            self._storage = self._project.create_storage(self.name)
        return self._storage

    def _set_output(self, data):
        self._output = data["output"][0]

    def _setup_executor_callbacks(self):
        self._executor._run_machine.observe("ready", lambda _: self.store(self.storage))
        self._executor._run_machine.observe("finished", self._set_output)
        self._executor._run_machine.observe(
            "finished", lambda _: self.store(self.storage)
        )

        self._executor._run_machine.observe("ready", self._add_to_database)
        self._executor._run_machine.observe("running", self._update_status("running"))
        self._executor._run_machine.observe("collect", self._update_status("collect"))
        self._executor._run_machine.observe("finished", self._update_status("finished"))

    def run(
        self, executor: Union[Executor, str, None] = None
    ) -> Optional[ExecutionContext]:
        """
        Start execution of the job.

        If the job already has a database id and is not in "ready" state, do nothing.

        Args:
            executor (:class:`~.Executor`, str): specifies which executor to
                use, if `str` must be a method name of :class:`.ExecutorCreator`;
                if not given use the last created executor

        Returns:
            :class:`.Executor`: the executor that is running the task or nothing.
        """
        if (
            self._id is None
            or self.project.database.get_item(self.id).status == "ready"
        ):
            if executor is None:
                executor = "most_recent"
            if isinstance(executor, str):
                executor = getattr(self.project.create.executor, executor)()
            exe = self._executor = executor.submit(tasks=[self.task])
            self._setup_executor_callbacks()
            exe.run()
            return exe
        else:
            logging.info("Job already finished!")

    def wait(self, timeout: Optional[float] = None):
        """
        Wait until job is finished.

        Args:
            timeout (float, optional): maximum time to wait in seconds; wait
                indefinitely by default

        Raises:
            ValueError: if job status is not `finished` or `running`
        """
        if self.status == "finished":
            return
        if self.status != "running":
            raise ValueError("Job not running!")
        self._executor.wait(timeout=timeout)

    def remove(self):
        """
        Remove the job from the database and storage.

        Resets the internal state so that the job could be re-run from the same instance.
        """
        self.project.remove(self.id)
        self._id = None
        self._output = None
        self._storage = None
        self._executor = None

    def _add_to_database(self, _data):
        if self._id is None:
            entry = DatabaseEntry(
                name=self.name,
                project=self.project.path,
                username=state.settings.login_user,
                status="ready",
                jobtype=self.jobtype,
            )
            self._id = self.project.database.add_item(entry)
        return self.id

    def _update_status(self, status):
        return lambda _data: self.project.database.update_status(self.id, status)

    # Storable Impl'
    def _store(self, storage):
        # preferred solution, but not everything that can be pickled can go into HDF atm
        # self._executor.output[-1].store(storage, "output")
        storage["task"] = pickle_dump(self.task)
        if self._output is not None:
            storage["output"] = pickle_dump(self._output)

    def load(self, storage: GenericStorage = None):
        """
        Load job from storage and reload id from database.

        If `storage` is not given, use the default provided by the project.

        Args:
            storage (:class:`.GenericStorage`): where to read from
        """
        self._update_id()
        if storage is None:
            storage = self.storage
        self._task = pickle_load(storage["task"])
        # this would be correct, but since we pickle output and put it into a
        # HDF task it doesn't appear here yet!
        # if "output" in storage.list_groups():
        if "output" in storage.list_nodes():
            self._output = pickle_load(storage["output"])

    @classmethod
    def _restore(cls, storage, version):
        task = pickle_load(storage["task"])
        job = cls(task=task, project=storage.project, job_name=storage.name)
        job.load(storage=storage)
        return job



from concurrent.futures import Future, ProcessPoolExecutor
from concurrent.futures.process import _global_shutdown, _WorkItem, BrokenProcessPool
from sys import version_info

import cloudpickle


class CloudLoadsFuture(Future):
    def result(self, timeout=None):
        result = super().result(timeout=timeout)
        if isinstance(result, bytes):
            result = cloudpickle.loads(result)
        return result


class _CloudPickledCallable:
    def __init__(self, fnc: callable):
        self.fnc_serial = cloudpickle.dumps(fnc)

    def __call__(self, /, dumped_args, dumped_kwargs):
        fnc = cloudpickle.loads(self.fnc_serial)
        args = cloudpickle.loads(dumped_args)
        kwargs = cloudpickle.loads(dumped_kwargs)
        return cloudpickle.dumps(fnc(*args, **kwargs))

    @classmethod
    def dumps(cls, stuff):
        return cloudpickle.dumps(stuff)


class CloudpickleProcessPoolExecutor(ProcessPoolExecutor):
    """
    This class wraps `concurrent.futures.ProcessPoolExecutor` such that the submitted
    callable, its arguments, and its return value are all pickled using `cloudpickle`.
    In this way, the executor extends support to all objects which are cloud-pickleable,
    e.g. dynamically defined or decorated classes.

    To accomplish this, the underlying `concurrent.futures.Future` class used is
    replaced with our `CloudLoadsFuture`, which is identical except that calls to
    `result()` will first try to `cloudpickle.loads` and `bytes` results found.

    Examples:
        Consider a class created from a function dynamically with a decorator.
        These are not normally pickleable, so in this example we should how this class
        allows us to submit a method from such a class, that both takes as an argument
        and returns such an unpickleable class.
        Actions such as registering callbacks and waiting for results behave just like
        normal.
        >>> from functools import partialmethod
        >>>
        >>> from pyiron_contrib.executors import CloudpickleProcessPoolExecutor
        >>>
        >>> class Foo:
        ...     '''
        ...     A base class to be dynamically modified for testing our executor.
        ...     '''
        ...     def __init__(self, fnc: callable):
        ...         self.fnc = fnc
        ...         self.result = None
        ...
        ...     @property
        ...     def run(self):
        ...         return self.fnc
        ...
        ...     def process_result(self, future):
        ...         self.result = future.result()
        >>>
        >>>
        >>> def dynamic_foo():
        ...     '''
        ...     A decorator for dynamically modifying the Foo class.
        ...
        ...     Overrides the `fnc` input of `Foo` with the decorated function.
        ...     '''
        ...     def as_dynamic_foo(fnc: callable):
        ...         return type(
        ...             "DynamicFoo",
        ...             (Foo,),  # Define parentage
        ...             {
        ...                 "__init__": partialmethod(
        ...                     Foo.__init__,
        ...                     fnc
        ...                 )
        ...             },
        ...         )
        ...
        ...     return as_dynamic_foo
        >>>
        >>> @dynamic_foo()
        >>> def UnpicklableCallable(unpicklable_arg):
        ...     unpicklable_arg.result = "This was an arg"
        ...     return unpicklable_arg
        >>>
        >>>
        >>> instance = UnpicklableCallable()
        >>> arg = UnpicklableCallable()
        >>> executor = CloudpickleProcessPoolExecutor()
        >>> fs = executor.submit(instance.run, arg)
        >>> fs.add_done_callback(instance.process_result)
        >>> print(fs.done())
        False

        >>> print(fs.result().__class__.__name__)
        DynamicFoo

        >>> print(fs.done())
        True

        >>> print(instance.result.result)
        This was an arg
    """

    def submit(self, fn, /, *args, **kwargs):
        return self._submit(
            _CloudPickledCallable(fn),
            _CloudPickledCallable.dumps(args),
            _CloudPickledCallable.dumps(kwargs),
        )

    submit.__doc__ = ProcessPoolExecutor.submit.__doc__

    def _submit(self, fn, /, *args, **kwargs):
        """
        We override the regular `concurrent.futures.ProcessPoolExecutor` to use our
        custom future that unpacks cloudpickled results.

        This approach is simple, but the brute-force nature of it means we manually
        accommodate different implementations of `ProcessPoolExecutor` in different
        python versions.
        """
        if version_info.major != 3:
            raise RuntimeError(
                f"{self.__class__} is only built for python3, but got "
                f"{version_info.major}"
            )

        if version_info.minor == 8:
            return self._submit_3_8(fn, *args, **kwargs)
        elif version_info.minor >= 9:
            return self._submit_3_gt9(fn, *args, **kwargs)
        else:
            raise RuntimeError(
                f"{self.__class__} is only built for python 3.8+, but got "
                f"{version_info.major}.{version_info.minor}."
            )

    def _submit_3_gt9(self, fn, /, *args, **kwargs):
        with self._shutdown_lock:
            if self._broken:
                raise BrokenProcessPool(self._broken)
            if self._shutdown_thread:
                raise RuntimeError("cannot schedule new futures after shutdown")
            if _global_shutdown:
                raise RuntimeError(
                    "cannot schedule new futures after " "interpreter shutdown"
                )

            f = CloudLoadsFuture()
            w = _WorkItem(f, fn, args, kwargs)

            self._pending_work_items[self._queue_count] = w
            self._work_ids.put(self._queue_count)
            self._queue_count += 1
            # Wake up queue management thread
            self._executor_manager_thread_wakeup.wakeup()

            if self._safe_to_dynamically_spawn_children:
                self._adjust_process_count()
            self._start_executor_manager_thread()
            return f

    def _submit_3_8(*args, **kwargs):
        if len(args) >= 2:
            self, fn, *args = args
        elif not args:
            raise TypeError(
                "descriptor 'submit' of 'ProcessPoolExecutor' object "
                "needs an argument"
            )
        elif "fn" in kwargs:
            fn = kwargs.pop("fn")
            self, *args = args
            import warnings

            warnings.warn(
                "Passing 'fn' as keyword argument is deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            raise TypeError(
                "submit expected at least 1 positional argument, "
                "got %d" % (len(args) - 1)
            )

        with self._shutdown_lock:
            if self._broken:
                raise BrokenProcessPool(self._broken)
            if self._shutdown_thread:
                raise RuntimeError("cannot schedule new futures after shutdown")
            if _global_shutdown:
                raise RuntimeError(
                    "cannot schedule new futures after " "interpreter shutdown"
                )

            f = CloudLoadsFuture()
            w = _WorkItem(f, fn, args, kwargs)

            self._pending_work_items[self._queue_count] = w
            self._work_ids.put(self._queue_count)
            self._queue_count += 1
            # Wake up queue management thread
            self._queue_management_thread_wakeup.wakeup()

            self._start_queue_management_thread()
            return f



from __future__ import annotations

from typing import Optional

import numpy as np
from matplotlib import pyplot as plt

from pyiron_contrib.workflow.function import single_value_node


@single_value_node(output_labels="fig")
def scatter(
    x: Optional[list | np.ndarray] = None, y: Optional[list | np.ndarray] = None
):
    return plt.scatter(x, y)


nodes = [
    scatter,
]



# coding: utf-8
# Copyright (c) ICAMS, Ruhr University Bochum, 2022

## Executable required: $pyiron/resources/pacemaker/bin/run_pacemaker_tf_cpu.sh AND  run_pacemaker_tf.sh

import logging
from typing import List

import numpy as np
import os
import pandas as pd
import ruamel.yaml as yaml

from shutil import copyfile

from pyiron_base import (
    GenericJob,
    GenericParameters,
    state,
    Executable,
    FlattenedStorage,
)

from pyiron_contrib.atomistics.atomistics.job.trainingcontainer import (
    TrainingStorage,
    TrainingContainer,
)
from pyiron_contrib.atomistics.ml.potentialfit import PotentialFit

from pyiron_atomistics.atomistics.structure.atoms import (
    Atoms as pyironAtoms,
    ase_to_pyiron,
)
from ase.atoms import Atoms as aseAtoms

s = state.settings


class PacemakerJob(GenericJob, PotentialFit):
    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self.__name__ = "Pacemaker2022"
        self.__version__ = "0.2"

        self._train_job_id_list = []

        self.input = GenericParameters(table_name="input")
        self._cutoff = 7.0
        self.input["cutoff"] = self._cutoff
        self.input["metadata"] = {"comment": "pyiron-generated fitting job"}

        # data_config
        self.input["data"] = {}
        # potential_config
        self.input["potential"] = {
            "elements": [],
            "bonds": {
                "ALL": {
                    "radbase": "SBessel",
                    "rcut": self._cutoff,
                    "dcut": 0.01,
                    "radparameters": [5.25],
                }
            },
            "embeddings": {
                "ALL": {
                    "fs_parameters": [1, 1, 1, 0.5],
                    "ndensity": 2,
                    "npot": "FinnisSinclairShiftedScaled",
                }
            },
            "functions": {
                "ALL": {
                    "nradmax_by_orders": [15, 3, 2, 1],
                    "lmax_by_orders": [0, 3, 2, 1],
                }
            },
        }

        # fit_config
        self.input["fit"] = {
            "loss": {
                "L1_coeffs": 1e-8,
                "L2_coeffs": 1e-8,
                "kappa": 0.3,
                "w0_rad": 0,
                "w1_rad": 0,
                "w2_rad": 0,
            },
            "maxiter": 1000,
            "optimizer": "BFGS",
            "fit_cycles": 1,
        }
        self.input["backend"] = {
            "batch_size": 100,
            "display_step": 50,
            "evaluator": "tensorpot",
        }  # backend_config

        self.structure_data = None
        self._executable = None
        self._executable_activate()

        state.publications.add(self.publication)

    @property
    def elements(self):
        return self.input["potential"].get("elements")

    @elements.setter
    def elements(self, val):
        self.input["potential"]["elements"] = val

    @property
    def cutoff(self):
        return self._cutoff

    @cutoff.setter
    def cutoff(self, val):
        self._cutoff = val
        self.input["cutoff"] = self._cutoff
        self.input["potential"]["bonds"]["ALL"]["rcut"] = self._cutoff

    @property
    def publication(self):
        return {
            "pacemaker": [
                {
                    "title": "Efficient parametrization of the atomic cluster expansion",
                    "journal": "Physical Review Materials",
                    "volume": "6",
                    "number": "1",
                    "year": "2022",
                    "doi": "10.1103/PhysRevMaterials.6.013804",
                    "url": "https://doi.org/10.1103/PhysRevMaterials.6.013804",
                    "author": [
                        "Anton Bochkarev",
                        "Yury Lysogorskiy",
                        "Sarath Menon",
                        "Minaam Qamar",
                        "Matous Mrovec",
                        "Ralf Drautz",
                    ],
                },
                {
                    "title": "Performant implementation of the atomic cluster expansion (PACE) and application to copper and silicon",
                    "journal": "npj Computational Materials",
                    "volume": "7",
                    "number": "1",
                    "year": "2021",
                    "doi": "10.1038/s41524-021-00559-9",
                    "url": "https://doi.org/10.1038/s41524-021-00559-9",
                    "author": [
                        "Yury Lysogorskiy",
                        "Cas van der Oord",
                        "Anton Bochkarev",
                        "Sarath Menon",
                        "Matteo Rinaldi",
                        "Thomas Hammerschmidt",
                        "Matous Mrovec",
                        "Aidan Thompson",
                        "Gábor Csányi",
                        "Christoph Ortner",
                        "Ralf Drautz",
                    ],
                },
                {
                    "title": "Atomic cluster expansion for accurate and transferable interatomic potentials",
                    "journal": "Physical Review B",
                    "volume": "99",
                    "year": "2019",
                    "doi": "10.1103/PhysRevB.99.014104",
                    "url": "https://doi.org/10.1103/PhysRevB.99.014104",
                    "author": ["Ralf Drautz"],
                },
            ]
        }

    def _save_structure_dataframe_pckl_gzip(self, df):
        if "NUMBER_OF_ATOMS" not in df.columns and "number_of_atoms" in df.columns:
            df.rename(columns={"number_of_atoms": "NUMBER_OF_ATOMS"}, inplace=True)
        df["NUMBER_OF_ATOMS"] = df["NUMBER_OF_ATOMS"].astype(int)

        # TODO: reference energy subtraction ?
        if "energy_corrected" not in df.columns and "energy" in df.columns:
            df.rename(columns={"energy": "energy_corrected"}, inplace=True)

        if "atoms" in df.columns:
            # check if this is pyironAtoms  -> aseAtoms
            at = df.iloc[0]["atoms"]
            if isinstance(at, pyironAtoms):
                df["ase_atoms"] = df["atoms"].map(lambda s: s.to_ase())
                df.drop(columns=["atoms"], inplace=True)
            else:
                assert isinstance(
                    at, aseAtoms
                ), "'atoms' column is not a valid ASE Atoms object"
                df.rename(columns={"atoms": "ase_atom"}, inplace=True)
        elif "ase_atoms" not in df.columns:
            raise ValueError(
                "DataFrame should contain 'atoms' (pyiron Atoms) or 'ase_atoms' (ASE atoms) columns"
            )

        if "stress" in df.columns:
            df.drop(columns=["stress"], inplace=True)

        if "pbc" not in df.columns:
            df["pbc"] = df["ase_atoms"].map(lambda atoms: np.all(atoms.pbc))

        data_file_name = os.path.join(self.working_directory, "df_fit.pckl.gzip")
        logging.info(
            "Saving training structures dataframe into {} with pickle protocol = 4, compression = gzip".format(
                data_file_name
            )
        )
        df.to_pickle(data_file_name, compression="gzip", protocol=4)
        return data_file_name

    def write_input(self):
        # prepare datafile
        if self._train_job_id_list and self.structure_data is None:
            train_df = self.create_training_dataframe(self._train_job_id_list)
            self.structure_data = train_df

        if isinstance(self.structure_data, pd.DataFrame):
            logging.info("structure_data is pandas.DataFrame")
            data_file_name = self._save_structure_dataframe_pckl_gzip(
                self.structure_data
            )
            self.input["data"] = {"filename": data_file_name}
            elements_set = set()
            for at in self.structure_data["ase_atoms"]:
                elements_set.update(at.get_chemical_symbols())
            elements = sorted(elements_set)
            print("Set automatically determined list of elements: {}".format(elements))
            self.elements = elements
        elif isinstance(self.structure_data, str):  # filename
            if os.path.isfile(self.structure_data):
                logging.info("structure_data is valid file path")
                self.input["data"] = {"filename": self.structure_data}
            else:
                raise ValueError(
                    "Provided structure_data filename ({}) doesn't exists".format(
                        self.structure_data
                    )
                )
        elif hasattr(
            self.structure_data, "get_pandas"
        ):  # duck-typing check for TrainingContainer
            logging.info("structure_data is TrainingContainer")
            df = self.structure_data.to_pandas()
            data_file_name = self._save_structure_dataframe_pckl_gzip(df)
            self.input["data"] = {"filename": data_file_name}
        elif self.structure_data is None:
            raise ValueError(
                "`structure_data` is none, but should be pd.DataFrame, TrainingContainer or valid pickle.gzip filename"
            )

        metadata_dict = self.input["metadata"]
        metadata_dict["pyiron_job_id"] = str(self.job_id)

        input_yaml_dict = {
            "cutoff": self.input["cutoff"],
            "metadata": metadata_dict,
            "potential": self.input["potential"],
            "data": self.input["data"],
            "fit": self.input["fit"],
            "backend": self.input["backend"],
        }

        if isinstance(self.input["potential"], str):
            pot_file_name = self.input["potential"]
            if os.path.isfile(pot_file_name):
                logging.info("Input potential is filename")
                pot_basename = os.path.basename(pot_file_name)
                copyfile(
                    pot_file_name, os.path.join(self.working_directory, pot_basename)
                )
                input_yaml_dict["potential"] = pot_basename
            else:
                raise ValueError(
                    "Provided potential filename ({}) doesn't exists".format(
                        self.input["potential"]
                    )
                )

        with open(os.path.join(self.working_directory, "input.yaml"), "w") as f:
            yaml.dump(input_yaml_dict, f)

    def _analyse_log(self, logfile="metrics.txt"):
        metrics_filename = os.path.join(self.working_directory, logfile)

        metrics_df = pd.read_csv(metrics_filename, sep="\s+")
        res_dict = metrics_df.to_dict(orient="list")
        return res_dict

    def collect_output(self):
        final_potential_filename_yaml = self.get_final_potential_filename()
        with open(final_potential_filename_yaml, "r") as f:
            yaml_lines = f.readlines()
        final_potential_yaml_string = "".join(yaml_lines)

        with open(self.get_final_potential_filename_ace(), "r") as f:
            ace_lines = f.readlines()
        final_potential_yace_string = "".join(ace_lines)

        with open(self.get_final_potential_filename_ace(), "r") as f:
            yace_data = yaml.safe_load(f)

        elements_name = yace_data["elements"]

        with self.project_hdf5.open("output/potential") as h5out:
            h5out["yaml"] = final_potential_yaml_string
            h5out["yace"] = final_potential_yace_string
            h5out["elements_name"] = elements_name

        log_res_dict = self._analyse_log()

        with self.project_hdf5.open("output/log") as h5out:
            for key, arr in log_res_dict.items():
                h5out[key] = arr

        # training data
        training_data_fname = os.path.join(
            self.working_directory, "fitting_data_info.pckl.gzip"
        )
        df = pd.read_pickle(training_data_fname, compression="gzip")
        df["atoms"] = df.ase_atoms.map(ase_to_pyiron)
        training_data_ts = TrainingStorage()
        for _, r in df.iterrows():
            training_data_ts.add_structure(
                r.atoms,
                energy=r.energy_corrected,
                forces=r.forces,
                identifier=r["name"],
            )

        # predicted data
        predicted_fname = os.path.join(self.working_directory, "train_pred.pckl.gzip")
        df = pd.read_pickle(predicted_fname, compression="gzip")
        predicted_data_fs = FlattenedStorage()
        predicted_data_fs.add_array("energy", dtype=np.float64, shape=(), per="chunk")
        predicted_data_fs.add_array(
            "energy_true", dtype=np.float64, shape=(), per="chunk"
        )

        predicted_data_fs.add_array(
            "number_of_atoms", dtype=np.int64, shape=(), per="chunk"
        )

        predicted_data_fs.add_array(
            "forces", dtype=np.float64, shape=(3,), per="element"
        )
        predicted_data_fs.add_array(
            "forces_true", dtype=np.float64, shape=(3,), per="element"
        )
        for i, r in df.iterrows():
            identifier = r["name"] if "name" in r else str(i)
            predicted_data_fs.add_chunk(
                r["NUMBER_OF_ATOMS"],
                identifier=identifier,
                energy=r.energy_pred,
                forces=r.forces_pred,
                energy_true=r.energy_corrected,
                forces_true=r.forces,
                number_of_atoms=r.NUMBER_OF_ATOMS,
                energy_per_atom=r.energy_pred / r.NUMBER_OF_ATOMS,
                energy_per_atom_true=r.energy_corrected / r.NUMBER_OF_ATOMS,
            )

        with self.project_hdf5.open("output") as hdf5_output:
            training_data_ts.to_hdf(hdf=hdf5_output, group_name="training_data")
            predicted_data_fs.to_hdf(hdf=hdf5_output, group_name="predicted_data")

    def get_lammps_potential(self):
        elements_name = self["output/potential/elements_name"]
        elem = " ".join(elements_name)
        pot_file_name = self.get_final_potential_filename_ace()
        pot_dict = {
            "Config": [
                [
                    "pair_style pace\n",
                    "pair_coeff  * * {} {}\n".format(pot_file_name, elem),
                ]
            ],
            "Filename": [""],
            "Model": ["ACE"],
            "Name": [self.job_name],
            "Species": [elements_name],
        }

        ace_potential = pd.DataFrame(pot_dict)

        return ace_potential

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as h5in:
            self.input.to_hdf(h5in)

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as h5in:
            self.input.from_hdf(h5in)

    def get_final_potential_filename(self):
        return os.path.join(self.working_directory, "output_potential.yaml")

    def get_final_potential_filename_ace(self):
        return os.path.join(self.working_directory, "output_potential.yace")

    def get_current_potential_filename(self):
        return os.path.join(self.working_directory, "interim_potential_0.yaml")

    # To link to the executable from the notebook
    def _executable_activate(self, enforce=False, codename="pacemaker"):
        if self._executable is None or enforce:
            self._executable = Executable(
                codename=codename,
                module="pacemaker",
                path_binary_codes=state.settings.resource_paths,
            )

    def _add_training_data(self, container: TrainingContainer) -> None:
        self.add_job_to_fitting(container.id)

    def add_job_to_fitting(self, job_id, *args, **kwargs):
        self._train_job_id_list.append(job_id)

    def _get_training_data(self) -> TrainingStorage:
        return self["output/training_data"].to_object()

    def _get_predicted_data(self) -> FlattenedStorage:
        return self["output/predicted_data"].to_object()

    # copied/adapted from mlip.py
    def create_training_dataframe(
        self, _train_job_id_list: List = None
    ) -> pd.DataFrame:
        if _train_job_id_list is None:
            _train_job_id_list = self._train_job_id_list
        df_list = []
        for job_id in _train_job_id_list:
            ham = self.project.inspect(job_id)
            if ham.__name__ == "TrainingContainer":
                job = ham.to_object()
                data_df = job.to_pandas()
                df_list.append(data_df)
            else:
                raise NotImplementedError(
                    "Currently only TrainingContainer is supported"
                )

        total_training_df = pd.concat(df_list, axis=0)
        total_training_df.reset_index(drop=True, inplace=True)

        return total_training_df



__version__ = "0.1"
__all__ = []

import warnings

try:
    from pyiron import Project
except:
    warnings.warn("pyiron module not found, importing Project from pyiron_base")
    from pyiron_base import Project

from pyiron_base import JOB_CLASS_DICT
from pyiron_contrib.generic.storage_interface_toolkit import StorageInterfaceFactory

Project.register_tools("storage_interface", StorageInterfaceFactory)

# Make classes available for new pyiron version
JOB_CLASS_DICT.update(
    {
        "ProtoMinimGradDes": "pyiron_contrib.protocol.compound.minimize",
        "ProtoMD": "pyiron_contrib.protocol.compound.molecular_dynamics",
        "ProtoConfinedMD": "pyiron_contrib.protocol.compound.molecular_dynamics",
        "ProtoHarmMD": "pyiron_contrib.protocol.compound.molecular_dynamics",
        "ProtoNEBSer": "pyiron_contrib.protocol.compound.nudged_elastic_band",
        "ProtocolQMMM": "pyiron_contrib.protocol.compound.qmmm",
        "ProtoHarmTILDSer": "pyiron_contrib.protocol.compound.thermodynamic_integration",
        "ProtoHarmTILDPar": "pyiron_contrib.protocol.compound.thermodynamic_integration",
        "ProtoVacTILDSer": "pyiron_contrib.protocol.compound.thermodynamic_integration",
        "ProtoVacTILDPar": "pyiron_contrib.protocol.compound.thermodynamic_integration",
        "ProtoVacForm": "pyiron_contrib.protocol.compound.thermodynamic_integration",
        "ProtoFTSEvoSer": "pyiron_contrib.protocol.compound.finite_temperature_string",
        "ProtoFTSEvoPar": "pyiron_contrib.protocol.compound.finite_temperature_string",
        "ImageJob": "pyiron_contrib.image.job",
        "LangevinAse": "pyiron_contrib.atomistics.interactive.langevin",
        "Mixer": "pyiron_contrib.atomistics.interactive.mixer",
        "ParameterMaster": "pyiron_contrib.atomistics.dft.parametermaster",
        "MonteCarloMaster": "pyiron_contrib.atomistics.interactive.montecarlo",
        "RandSpg": "pyiron_contrib.atomistics.randspg.randspg",
        "Fenics": "pyiron_contrib.continuum.fenics.job.generic",
        "FenicsLinearElastic": "pyiron_contrib.continuum.fenics.job.elastic",
        "TrainingContainer": "pyiron_contrib.atomistics.atomistics.job.trainingcontainer",
        "RandomDisMaster": "pyiron_contrib.atomistics.mlip.masters",
        "RandomMDMaster": "pyiron_contrib.atomistics.mlip.masters",
        "RunnerFit": "pyiron_contrib.atomistics.runner.job",
        "MlipSelect": "pyiron_contrib.atomistics.mlip.mlipselect",
        "Mlip": "pyiron_contrib.atomistics.mlip.mlip",
        "LammpsMlip": "pyiron_contrib.atomistics.mlip.lammps",
        "MlipJob": "pyiron_contrib.atomistics.mlip.mlipjob",
        "Atomicrex": "pyiron_contrib.atomistics.atomicrex.atomicrex_job",
        "StructureMasterInt": "pyiron_contrib.atomistics.atomistics.job.structurelistmasterinteractive",
        "StorageJob": "pyiron_contrib.RDM.storagejob",
        "MlipDescriptors": "pyiron_contrib.atomistics.mlip.mlipdescriptors",
        "PacemakerJob": "pyiron_contrib.atomistics.pacemaker.job",
        "MeamFit": "pyiron_contrib.atomistics.meamfit.meamfit",
        "Cp2kJob": "pyiron_contrib.atomistics.cp2k.job",
        "PiMD": "pyiron_contrib.atomistics.ipi.ipi_jobs",
        "GleMD": "pyiron_contrib.atomistics.ipi.ipi_jobs",
        "PigletMD": "pyiron_contrib.atomistics.ipi.ipi_jobs",
        "LammpsInteractiveWithoutOutput": "pyiron_contrib.nofiles.lammps",
        "SQSJobWithoutOutput": "pyiron_contrib.nofiles.sqs",
        "ElasticMatrixJobWithoutFiles": "pyiron_contrib.nofiles.elastic",
        "MurnaghanWithoutFiles": "pyiron_contrib.nofiles.murn",
        "PhonopyJobWithoutFiles": "pyiron_contrib.nofiles.phonopy",
        "SQSMasterMPI": "pyiron_contrib.nofiles.master",
        "LAMMPSMinimizeMPI": "pyiron_contrib.nofiles.master",
        "LAMMPSElasticMPI": "pyiron_contrib.nofiles.master",
        "LAMMPSMinimizeElasticMPI": "pyiron_contrib.nofiles.master",
        "FitsnapJob": "pyiron_contrib.atomistics.fitsnap.job",
        "QuasiHarmonicApproximation": "pyiron_contrib.atomistics.atomistics.master.qha",
    }
)


from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



import numpy as np
from pybispectrum import calc_bispectrum_names
from fitsnap3lib.scrapers.ase_funcs import get_apre, create_shared_arrays


def ase_scraper(
    s, frames, energies, forces, stresses=[[0, 0, 0], [0, 0, 0], [0, 0, 0]]
):
    """
    Custom function to allocate shared arrays used in Calculator and build the internal list of
    dictionaries `data` of configuration info. Customized version of `fitsnap3lib.scrapers.ase_funcs`.

    Args:
        s: fitsnap instance.
        frames: list or array of ASE atoms objects.
        energies: array of energies.
        forces: array of forces for all configurations.
        stresses: array of stresses for all configurations.

    Creates a list of data dictionaries `s.data` suitable for fitsnap descriptor calculation.
    If running in parallel, this list will be distributed over procs, so that each proc will have a
    portion of the list.
    """

    create_shared_arrays(s, frames)
    s.data = [
        collate_data(a, e, f, s)
        for (a, e, f, s) in zip(frames, energies, forces, stresses)
    ]


def collate_data(atoms, energy, forces, stresses):
    """
    Function to organize fitting data for FitSNAP from ASE atoms objects.

    Args:
        atoms: ASE atoms object for a single configuration of atoms.
        energy: energy of a configuration.
        forces: numpy array of forces for a configuration.
        stresses: numpy array of stresses for a configuration.

    Returns a fitsnap data dictionary for a single configuration.
    """

    # make a data dictionary for this config

    apre = get_apre(cell=atoms.cell)
    R = np.dot(np.linalg.inv(atoms.cell), apre)

    positions = np.matmul(atoms.get_positions(), R)
    cell = apre.T

    data = {}
    data["PositionsStyle"] = "angstrom"
    data["AtomTypeStyle"] = "chemicalsymbol"
    data["StressStyle"] = "bar"
    data["LatticeStyle"] = "angstrom"
    data["EnergyStyle"] = "electronvolt"
    data["ForcesStyle"] = "electronvoltperangstrom"
    data["Group"] = "Displaced_BCC"
    data["File"] = None
    data["Stress"] = stresses
    data["Positions"] = positions
    data["Energy"] = energy
    data["AtomTypes"] = atoms.get_chemical_symbols()
    data["NumAtoms"] = len(atoms)
    data["Forces"] = forces
    data["QMLattice"] = cell
    data["test_bool"] = 0
    data["Lattice"] = cell
    data["Rotation"] = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    data["Translation"] = np.zeros((len(atoms), 3))
    data["eweight"] = 1.0
    data["fweight"] = 1.0 / 150.0
    data["vweight"] = 0.0

    return data


def subsample_twojmax(total_bispect, twojmax_lst):
    bi_spect_names_str_lst = [str(lst) for lst in total_bispect]
    twojmax_master_str_lst = [
        [str(lst) for lst in calc_bispectrum_names(twojmax=tjm)] for tjm in twojmax_lst
    ]
    ind_lst = [
        [desc in desc_lst for desc in bi_spect_names_str_lst]
        for desc_lst in twojmax_master_str_lst
    ]
    return ind_lst



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyiron_contrib.workflow.channels import Channel


class HasChannel(ABC):
    """
    A mix-in class for use with the `Channel` class.
    A `Channel` is able to (attempt to) connect to any child instance of `HasConnection`
    by looking at its `connection` attribute.

    This is useful for letting channels attempt to connect to non-channel objects
    directly by pointing them to some channel that object holds.
    """

    @property
    @abstractmethod
    def channel(self) -> Channel:
        pass



class DotDict(dict):
    def __getattr__(self, item):
        return self.__getitem__(item)

    def __setattr__(self, key, value):
        self[key] = value

    def __dir__(self):
        return set(super().__dir__() + list(self.keys()))



from pathlib import Path


def delete_files_and_directories_recursively(path):
    if not path.exists():
        return
    for item in path.rglob("*"):
        if item.is_file():
            item.unlink()
        else:
            delete_files_and_directories_recursively(item)
    path.rmdir()


def categorize_folder_items(folder_path):
    types = [
        "dir",
        "file",
        "mount",
        "symlink",
        "block_device",
        "char_device",
        "fifo",
        "socket",
    ]
    results = {t: [] for t in types}

    for item in folder_path.iterdir():
        for tt in types:
            try:
                if getattr(item, f"is_{tt}")():
                    results[tt].append(str(item))
            except NotImplementedError:
                pass
    return results


class DirectoryObject:
    def __init__(self, directory):
        self.path = Path(directory)
        self.create()

    def create(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def delete(self):
        delete_files_and_directories_recursively(self.path)

    def list_content(self):
        return categorize_folder_items(self.path)

    def __len__(self):
        return sum([len(cc) for cc in self.list_content().values()])

    def __repr__(self):
        return f"DirectoryObject(directory='{self.path}')\n{self.list_content()}"

    def get_path(self, file_name):
        return self.path / file_name

    def file_exists(self, file_name):
        return self.get_path(file_name).is_file()

    def write(self, file_name, content, mode="w"):
        with self.get_path(file_name).open(mode=mode) as f:
            f.write(content)

    def create_subdirectory(self, path):
        return DirectoryObject(self.path / path)

    def create_file(self, file_name):
        return FileObject(file_name, self)


class FileObject:
    def __init__(self, file_name: str, directory: DirectoryObject):
        self.directory = directory
        self._file_name = file_name

    @property
    def file_name(self):
        return self._file_name

    @property
    def path(self):
        return self.directory.path / Path(self._file_name)

    def write(self, content, mode="x"):
        self.directory.write(file_name=self.file_name, content=content, mode=mode)

    def read(self, mode="r"):
        with open(self.path, mode=mode) as f:
            return f.read()

    def is_file(self):
        return self.directory.file_exists(self.file_name)

    def delete(self):
        self.path.unlink()



import abc
import importlib
from typing import Any, Union, Optional

from pyiron_base import DataContainer
from pyiron_contrib.tinybase import __version__ as base__version__

import pickle
import codecs


# utility functions until ASE can be HDF'd
def pickle_dump(obj):
    return codecs.encode(pickle.dumps(obj), "base64").decode()


def pickle_load(buf):
    return pickle.loads(codecs.decode(buf.encode(), "base64"))


class GenericStorage(abc.ABC):
    """
    Generic interface to store things.

    The canonical implementation and model is ProjectHDFio from base, but other implementations are thinkable (S3, in
    memory, etc.)

    The concepts are borrowed mostly from HDF5 files.  There are groups, :meth:`.list_groups()`, and nodes,
    :meth:`.list_nodes()`, inside any storage.  Every group has a name, :attr:`.name`.

    Implementations must allow multiple objects of this class to refer to the same underlying storage group at the same
    time and access via the methods here must be atomic.
    """

    @abc.abstractmethod
    def __getitem__(self, item: str) -> Union["GenericStorage", Any]:
        """
        Return a value from storage.

        If `item` is in :meth:`.list_groups()` this must return another :class:`.GenericStorage`.

        Args:
            item (str): name of value

        Returns:
            :class:`.GenericStorage`: if `item` refers to a sub group
            object: value that is stored under `item`

        Raises:
            KeyError: `item` is neither a node or a sub group of this group
        """
        pass

    @abc.abstractmethod
    def __setitem__(self, item: str, value: Any):
        """
        Set a value to storage.

        Args:
            item (str): name of the value
            value (object): value to store
        """
        pass

    @abc.abstractmethod
    def create_group(self, name):
        """
        Create a new sub group.

        Args:
            name (str): name of the new group
        """
        pass

    def open(self, name: str) -> "GenericStorage":
        """
        Descend into a sub group.

        If `name` does not exist yet, create a new group.  Calling :meth:`~.close` on the returned object returns this
        object.

        Args:
            name (str): name of sub group

        Returns:
            :class:`.GenericStorage`: sub group
        """
        # FIXME: what if name in self.list_nodes()
        new = self.create_group(name)
        new._prev = self
        return new

    def close(self) -> "GenericStorage":
        """
        Surface from a sub group.

        If this object was not returned from a previous call to :meth:`.open` it returns itself silently.
        """
        try:
            return self._prev
        except AttributeError:
            return self

    @abc.abstractmethod
    def list_nodes(self) -> list[str]:
        """
        List names of values inside group.
        """
        pass

    @abc.abstractmethod
    def list_groups(self) -> list[str]:
        """
        List name of sub groups.
        """
        pass

    # DESIGN: this mostly exists to help to_object()ing GenericTinyJob, but it introduces a circular-ish connection.
    # Maybe there's another way to do it?
    @property
    @abc.abstractmethod
    def project(self):
        """
        The project that this storage belongs to.
        """
        pass

    @property
    @abc.abstractmethod
    def name(self):
        """
        The name of the group that this object is currently pointing to.

        (In ProjectHDFio speak, this is the os.path.basename of h5_path)
        """
        pass

    def to_object(self) -> "Storable":
        """
        Instantiate an object serialized to this group.

        Returns:
            :class:`.Storable`: deserialized object

        Raises:
            ValueError: no was object serialized in this group (either NAME, MODULE or VERSION values missing)
            RuntimeError: failed to import serialized object (probably some libraries not installed)
        """
        try:
            name = self["NAME"]
            module = self["MODULE"]
            version = self["VERSION"]
        except KeyError:
            raise ValueError("No object was serialized in this group!")
        try:
            cls = getattr(importlib.import_module(module), name)
        except ImportError:
            raise RuntimeError("Failed to import serialized object!")
        return cls.restore(self, version=version)


class ProjectHDFioStorageAdapter(GenericStorage):
    """
    Adapter class around ProjectHDFio to let it be used as a GenericStorage.
    """

    def __init__(self, project, hdf):
        self._project = project
        self._hdf = hdf

    def __getitem__(self, item):
        return self._hdf[item]

    def __setitem__(self, item, value):
        self._hdf[item] = value

    def create_group(self, name):
        return ProjectHDFioStorageAdapter(self._project, self._hdf.create_group(name))

    def list_nodes(self):
        return self._hdf.list_nodes()

    def list_groups(self):
        return self._hdf.list_groups()

    # compat with small bug in base ProjectHDFio
    list_dirs = list_groups

    @property
    def project(self):
        return self._project

    @property
    def name(self):
        return self._hdf.name


class DataContainerAdapter(GenericStorage):
    """
    Provides in memory location to store objects.
    """

    def __init__(self, project, cont: DataContainer, name):
        self._project = project
        self._cont = cont
        self._name = name

    def __getitem__(self, item: str) -> Union["GenericStorage", Any]:
        v = self._cont[item]
        if isinstance(v, DataContainer):
            return self.__class__(self._project, v, item)
        else:
            return v

    def __setitem__(self, item: str, value: Any):
        self._cont[item] = value

    def create_group(self, name):
        if name not in self._cont:
            d = self._cont.create_group(name)
        else:
            d = self._cont[name]
        return self.__class__(self._project, d, name)

    def list_nodes(self):
        return self._cont.list_nodes()

    def list_groups(self):
        return self._cont.list_groups()

    @property
    def project(self):
        return self._project

    @property
    def name(self):
        return self._name


# DESIGN: equivalent of HasHDF but with generalized language
class Storable(abc.ABC):
    """
    Interface for classes that can store themselves to a :class:`~.GenericStorage`

    Necessary overrides are :meth:`._store` and :meth:`._restore`.
    """

    @abc.abstractmethod
    def _store(self, storage):
        pass

    def _store_type(self, storage):
        storage["NAME"] = self.__class__.__name__
        storage["MODULE"] = self.__class__.__module__
        # DESIGN: what happens with objects defined in different parts of pyiron{_contrib,_atomistics,_base}?  Which
        # version do we save?
        storage["VERSION"] = base__version__

    def store(self, storage: GenericStorage, group_name: Optional[str] = None):
        """
        Store object into storage.

        Args:
            storage (:class:`.GenericStorage`): storage to write to
            group_name (str): if given descend into this subgroup first
        """
        if group_name is not None:
            storage = storage.create_group(group_name)
        self._store_type(storage)
        self._store(storage)

    @classmethod
    @abc.abstractmethod
    def _restore(cls, storage: GenericStorage, version: str) -> "Storable":
        pass

    @classmethod
    def restore(cls, storage: GenericStorage, version: str) -> "Storable":
        """
        Restore an object of type `cls` from storage.

        Args:
            storage (:class:`.GenericStorage`): storage to read from
            version (str): version string of pyiron that wrote the object

        Return:
            :class:`.Storable`: deserialized object

        Raises:
            ValueError: failed to restore object
        """
        try:
            return cls._restore(storage, version)
        except Exception as e:
            raise ValueError(f"Failed to restore object with {e}")


class HasHDFAdapaterMixin(Storable):
    """
    Implements :class:`.Storable` in terms of HasHDF.  Make any sub class of it a subclass :class:`.Storable` as well by
    mixing this class in.
    """

    def _store(self, storage):
        self._to_hdf(storage)

    @classmethod
    def _restore(cls, storage, version):
        kw = cls.from_hdf_args(storage)
        obj = cls(**kw)
        obj._from_hdf(storage, version)
        return obj



"""
Provides the main workhorse class for creating and running workflows.

This class is intended as the single point of entry for users making an import.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pyiron_contrib.workflow.composite import Composite
from pyiron_contrib.workflow.io import Inputs, Outputs


if TYPE_CHECKING:
    from pyiron_contrib.workflow.node import Node


class Workflow(Composite):
    """
    Workflows are a dynamic composite node -- i.e. they hold and run a collection of
    nodes (a subgraph) which can be dynamically modified (adding and removing nodes,
    and modifying their connections).

    Nodes can be added to the workflow at instantiation or with dot-assignment later on.
    They are then accessible either under the `nodes` dot-dictionary, or just directly
    by dot-access on the workflow object itself.

    Using the `input` and `output` attributes, the workflow gives access to all the
    IO channels among its nodes which are currently unconnected.

    Examples:
        We allow adding nodes to workflows in five equivalent ways:
        >>> from pyiron_contrib.workflow.workflow import Workflow
        >>> from pyiron_contrib.workflow.function import Function
        >>>
        >>> def fnc(x=0):
        ...     return x + 1
        >>>
        >>> n1 = Function(fnc, label="n1")
        >>>
        >>> wf = Workflow("my_workflow", n1)  # As *args at instantiation
        >>> wf.add(Function(fnc, label="n2"))  # Passing a node to the add caller
        >>> wf.add.Function(fnc, label="n3")  # Instantiating from add
        >>> wf.n4 = Function(fnc, label="whatever_n4_gets_used")
        >>> # By attribute assignment
        >>> Function(fnc, label="n5", parent=wf)
        >>> # By instantiating the node with a workflow

        By default, the node naming scheme is strict, so if you try to add a node to a
        label that already exists, you will get an error. This behaviour can be changed
        at instantiation with the `strict_naming` kwarg, or afterwards by assigning a
        bool to this property. When deactivated, repeated assignments to the same label
        just get appended with an index:
        >>> wf.strict_naming = False
        >>> wf.my_node = Function(fnc, x=0)
        >>> wf.my_node = Function(fnc, x=1)
        >>> wf.my_node = Function(fnc, x=2)
        >>> print(wf.my_node.inputs.x, wf.my_node0.inputs.x, wf.my_node1.inputs.x)
        0, 1, 2

        The `Workflow` class is designed as a single point of entry for workflows, so
        you can also access decorators to define new node classes right from the
        workflow (cf. the `Node` docs for more detail on the node types).
        Let's use these to explore a workflow's input and output, which are dynamically
        generated from the unconnected IO of its nodes:
        >>> @Workflow.wrap_as.function_node(output_labels="y")
        >>> def plus_one(x: int = 0):
        ...     return x + 1
        >>>
        >>> wf = Workflow("io_workflow")
        >>> wf.first = plus_one()
        >>> wf.second = plus_one()
        >>> print(len(wf.inputs), len(wf.outputs))
        2 2

        If we connect the output of one node to the input of the other, there are fewer
        dangling channels for the workflow IO to find:
        >>> wf.second.inputs.x = wf.first.outputs.y
        >>> print(len(wf.inputs), len(wf.outputs))
        1 1

        The workflow joins node lavels and channel labels with a `_` character to
        provide direct access to the output:
        >>> print(wf.outputs.second_y.value)
        2

        These input keys can be used when calling the workflow to update the input. In
        our example, the nodes update automatically when their input gets updated, so
        all we need to do to see updated workflow output is update the input:
        >>> out = wf(first_x=10)
        >>> out
        {'second_y': 12}

        Note: this _looks_ like a dictionary, but has some extra convenience that we
        can dot-access data:
        >>> out.second_y
        12

        Workflows also give access to packages of pre-built nodes under different
        namespaces, e.g.
        >>> wf = Workflow("with_prebuilt")
        >>>
        >>> wf.structure = wf.add.atomistics.BulkStructure(
        ...     repeat=3,
        ...     cubic=True,
        ...     element="Al"
        ... )
        >>> wf.engine = wf.add.atomistics.Lammps(structure=wf.structure)
        >>> wf.calc = wf.add.atomistics.CalcMd(
        ...     job=wf.engine,
        ...     run_on_updates=True,
        ...     update_on_instantiation=True,
        ... )
        >>> wf.plot = wf.add.standard.Scatter(
        ...     x=wf.calc.outputs.steps,
        ...     y=wf.calc.outputs.temperature
        ... )

    TODO: Workflows can be serialized.

    TODO: Once you're satisfied with how a workflow is structured, you can export it
        as a macro node for use in other workflows. (Maybe we should allow for nested
        workflows without exporting to a node? I was concerned then what happens to the
        nesting abstraction if, instead of accessing IO through the workflow's IO flags,
        a user manually connects IO from individual nodes from two different, nested or
        sibling workflows when those connections were _previously internal to their own
        workflow_. This seems very unsafe. Maybe there is something like a lock we can
        apply that falls short of a full export, but still guarantees the internal
        integrity of workflows when they're used somewhere else?
    """

    def __init__(
        self, label: str, *nodes: Node, run_on_updates: bool = True, strict_naming=True
    ):
        super().__init__(
            label=label,
            parent=None,
            run_on_updates=run_on_updates,
            strict_naming=strict_naming,
        )

        for node in nodes:
            self.add_node(node)

    @property
    def inputs(self) -> Inputs:
        inputs = Inputs()
        for node_label, node in self.nodes.items():
            for channel in node.inputs:
                if not channel.connected:
                    inputs[f"{node_label}_{channel.label}"] = channel
        return inputs

    @property
    def outputs(self) -> Outputs:
        outputs = Outputs()
        for node_label, node in self.nodes.items():
            for channel in node.outputs:
                if not channel.connected:
                    outputs[f"{node_label}_{channel.label}"] = channel
        return outputs

    def to_node(self):
        """
        Export the workflow to a macro node, with the currently exposed IO mapped to
        new IO channels, and the workflow mapped into the node_function.
        """
        raise NotImplementedError

    # (De)serialization is necessary throughout these classes, but not implemented here
    def serialize(self):
        raise NotImplementedError

    def deserialize(self, source):
        raise NotImplementedError

    @property
    def parent(self) -> None:
        return None

    @parent.setter
    def parent(self, new_parent: None):
        # Currently workflows are not allowed to have a parent -- maybe we want to
        # change our minds on this in the future? If we do, we can just expose `parent`
        # as a kwarg and roll back this private var/property/setter protection and let
        # the super call in init handle everything
        if new_parent is not None:
            raise TypeError(
                f"{self.__class__} may only take None as a parent but got "
                f"{type(new_parent)}"
            )




# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-08-02T15:40:06-0600",
 "dirty": true,
 "error": null,
 "full-revisionid": "1f99d246c76029c2e4e3f09f0000e382e4424e92",
 "version": "0.1.12"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from itertools import islice
import numbers
import numpy as np
import os
import pandas as pd
import posixpath
import scipy.constants
from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage
from pyiron_base import (
    state,
    GenericParameters,
    GenericJob,
    Executable,
    FlattenedStorage,
)
from pyiron_atomistics import ase_to_pyiron, Atoms
from pyiron_contrib.atomistics.ml.potentialfit import PotentialFit
from pyiron_contrib.atomistics.mlip.cfgs import savecfgs, loadcfgs, Cfg
from pyiron_contrib.atomistics.mlip.potential import MtpPotential

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


gpa_to_ev_ang = (
    1e22 / scipy.constants.physical_constants["joule-electron volt relationship"][0]
)


class Mlip(GenericJob, PotentialFit):
    def __init__(self, project, job_name):
        super(Mlip, self).__init__(project, job_name)
        self.__version__ = "0.1.0"
        self.__name__ = "Mlip"
        self._executable_activate()
        self._job_dict = {}
        self.input = MlipParameter()
        self._command_line = CommandLine()
        self._potential = MtpPotential()

    def _executable_activate(self, enforce=False):
        if self._executable is None or enforce:
            if len(self.__module__.split(".")) > 1:
                self._executable = Executable(
                    codename=self.__name__,
                    module=self.__module__.split(".")[-2],
                    path_binary_codes=state.settings.resource_paths,
                )
            else:
                self._executable = Executable(
                    codename=self.__name__,
                    path_binary_codes=state.settings.resource_paths,
                )

    @property
    def calculation_dataframe(self):
        (
            job_id_lst,
            time_step_start_lst,
            time_step_end_lst,
            time_step_delta_lst,
        ) = self._job_dict_lst(self._job_dict)
        return pd.DataFrame(
            {
                "Job id": job_id_lst,
                "Start config": time_step_start_lst,
                "End config": time_step_end_lst,
                "Step": time_step_delta_lst,
            }
        )

    @property
    def potential_files(self):
        pot = os.path.join(self.working_directory, "Trained.mtp_")
        states = os.path.join(self.working_directory, "state.mvs")
        if os.path.exists(pot) and os.path.exists(states):
            return [pot, states]

    def _get_elements(self):
        """
        Return elements in training in insertion order, i.e. elements seen earlier get lower indices.
        """
        elements = []
        for job_id in self._job_dict:
            j = self.project.inspect(job_id)
            if j["NAME"] == "TrainingContainer":
                candidates = j.to_object().get_elements()
            else:
                candidates = j["input/structure/species"]
            for e in candidates:
                if e not in elements:
                    elements.append(e)
        return elements

    def potential_dataframe(self, elements=None):
        """
        :class:`pandas.DataFrame`: potential dataframe for lammps jobs

        .. attention::

            The `elements` argument must be given for in any case for non-unary alloys, as the scrapping code from the
            training data is currently broken!

        Args:
            elements (list of str, optional): name of elements in this potential in the order of their indices in pyiron
                                              structures, if not given use elements seen in training data
        """
        if elements is None:
            elements = self.input["species"]
            if elements is None:
                elements = self._get_elements()  # AAAH

        if self.status.finished:
            return pd.DataFrame(
                {
                    "Name": ["".join(elements)],
                    "Filename": [self.potential_files],
                    "Model": ["Custom"],
                    "Species": [elements],
                    "Config": [["pair_style mlip mlip.ini\n", "pair_coeff * *\n"]],
                }
            )
        else:
            raise ValueError(
                f"Potential only available after job is finished, not {self.status}!"
            )

    @property
    def potential(self):
        if self.status.finished and self._potential is not None:
            return self._potential
        else:
            raise ValueError("potential only available on successfully finished jobs")

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        self.input.read_only = True
        self._command_line.read_only = True

    def add_job_to_fitting(
        self, job_id, time_step_start=0, time_step_end=-1, time_step_delta=10
    ):
        if time_step_end == -1:
            time_step_end = (
                np.shape(self.project.inspect(int(job_id))["output/generic/cells"])[0]
                - 1
            )
        self._job_dict[job_id] = {
            "time_step_start": time_step_start,
            "time_step_end": time_step_end,
            "time_step_delta": time_step_delta,
        }

    def write_input(self):
        restart_file_lst = [
            os.path.basename(f) for f in self._restart_file_list
        ] + list(self._restart_file_dict.values())
        if "testing.cfg" not in restart_file_lst:
            species = self._write_test_set(
                file_name="testing.cfg", cwd=self.working_directory
            )
            species_count = len(species)
            self.input["species"] = species
        elif self.input["filepath"] != "auto":
            species_count = 0
        else:
            raise ValueError("speciescount not set")
        if "training.cfg" not in restart_file_lst:
            self._write_training_set(
                file_name="training.cfg", cwd=self.working_directory
            )
        self._copy_potential(
            species_count=species_count,
            file_name="start.mtp",
            cwd=self.working_directory,
        )
        if self.version != "1.0.0":
            self._command_line.activate_postprocessing()
        self._command_line[0] = self._command_line[0].replace(
            "energy_auto", str(self.input["energy-weight"])
        )
        self._command_line[0] = self._command_line[0].replace(
            "force_auto", str(self.input["force-weight"])
        )
        self._command_line[0] = self._command_line[0].replace(
            "stress_auto", str(self.input["stress-weight"])
        )
        self._command_line[0] = self._command_line[0].replace(
            "iteration_auto", str(int(self.input["iteration"]))
        )
        self._command_line.write_file(file_name="mlip.sh", cwd=self.working_directory)

        species = np.array(species)
        input_store = StructureStorage()
        input_store.add_array("energy", dtype=np.float64, shape=(), per="chunk")
        input_store.add_array("forces", dtype=np.float64, shape=(3,), per="element")
        input_store.add_array("stress", dtype=np.float64, shape=(6,), per="chunk")
        for cfg in loadcfgs(os.path.join(self.working_directory, "training.cfg")):
            struct = Atoms(
                symbols=species[np.cast[np.int64](cfg.types)],
                positions=cfg.pos,
                cell=cfg.lat,
                pbc=[True] * 3,
            )  # HACK for pbc
            input_store.add_structure(
                struct,
                identifier=cfg.desc,
                energy=cfg.energy,
                forces=cfg.forces,
                stress=cfg.stresses,
            )

        with self.project_hdf5.open("input") as hdf5_input:
            input_store.to_hdf(hdf=hdf5_input, group_name="training_data")

    def collect_logfiles(self):
        pass

    def collect_output(self):
        file_name = os.path.join(self.working_directory, "diff.cfg")
        if os.path.exists(file_name):
            _, _, _, _, _, _, _, job_id_diff_lst, timestep_diff_lst = read_cgfs(
                file_name
            )
        else:
            job_id_diff_lst, timestep_diff_lst = [], []
        file_name = os.path.join(self.working_directory, "selected.cfg")
        if os.path.exists(file_name):
            (
                _,
                _,
                _,
                _,
                _,
                _,
                _,
                job_id_new_training_lst,
                timestep_new_training_lst,
            ) = read_cgfs(file_name)
        else:
            job_id_new_training_lst, timestep_new_training_lst = [], []
        file_name = os.path.join(self.working_directory, "grades.cfg")
        if os.path.exists(file_name):
            (
                _,
                _,
                _,
                _,
                _,
                _,
                grades_lst,
                job_id_grades_lst,
                timestep_grades_lst,
            ) = read_cgfs(file_name)
        else:
            grades_lst, job_id_grades_lst, timestep_grades_lst = [], [], []
        try:
            self._potential.load(os.path.join(self.working_directory, "Trained.mtp_"))
        except:
            self.logger.warn(
                "Failed to parse potential file! job.potential will not be available."
            )
            self._potential = None

        training_store = FlattenedStorage()
        training_store.add_array("energy", dtype=np.float64, shape=(), per="chunk")
        training_store.add_array("forces", dtype=np.float64, shape=(3,), per="element")
        training_store.add_array("stress", dtype=np.float64, shape=(6,), per="chunk")
        for cfg in loadcfgs(os.path.join(self.working_directory, "training_efs.cfg")):
            training_store.add_chunk(
                len(cfg.pos),
                identifier=cfg.desc,
                energy=cfg.energy,
                forces=cfg.forces,
                stress=cfg.stresses,
            )

        testing_store = FlattenedStorage()
        testing_store.add_array("energy", dtype=np.float64, shape=(), per="chunk")
        testing_store.add_array("forces", dtype=np.float64, shape=(3,), per="element")
        testing_store.add_array("stress", dtype=np.float64, shape=(6,), per="chunk")
        for cfg in loadcfgs(os.path.join(self.working_directory, "testing_efs.cfg")):
            testing_store.add_chunk(
                len(cfg.pos),
                identifier=cfg.desc,
                energy=cfg.energy,
                forces=cfg.forces,
                stress=cfg.stresses,
            )

        with self.project_hdf5.open("output") as hdf5_output:
            hdf5_output["grades"] = grades_lst
            hdf5_output["job_id"] = job_id_grades_lst
            hdf5_output["timestep"] = timestep_grades_lst
            hdf5_output["job_id_diff"] = job_id_diff_lst
            hdf5_output["timestep_diff"] = timestep_diff_lst
            hdf5_output["job_id_new"] = job_id_new_training_lst
            hdf5_output["timestep_new"] = timestep_new_training_lst
            if self._potential is not None:
                self._potential.to_hdf(hdf=hdf5_output)
            training_store.to_hdf(hdf=hdf5_output, group_name="training_efs")
            testing_store.to_hdf(hdf=hdf5_output, group_name="testing_efs")

    def get_structure(self, iteration_step=-1):
        job = self.project.load(self["output/job_id_diff"][iteration_step])
        return job.get_structure(self["output/timestep_diff"][iteration_step])

    def to_hdf(self, hdf=None, group_name=None):
        super(Mlip, self).to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(self._hdf5)
        with self._hdf5.open("input") as hdf_input:
            (
                job_id_lst,
                time_step_start_lst,
                time_step_end_lst,
                time_step_delta_lst,
            ) = self._job_dict_lst(self._job_dict)
            hdf_input["job_id_list"] = job_id_lst
            hdf_input["time_step_start_lst"] = time_step_start_lst
            hdf_input["time_step_end_lst"] = time_step_end_lst
            hdf_input["time_step_delta_lst"] = time_step_delta_lst

    def from_hdf(self, hdf=None, group_name=None):
        super(Mlip, self).from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(self._hdf5)
        with self._hdf5.open("input") as hdf_input:
            for job_id, start, end, delta in zip(
                hdf_input["job_id_list"],
                hdf_input["time_step_start_lst"],
                hdf_input["time_step_end_lst"],
                hdf_input["time_step_delta_lst"],
            ):
                self.add_job_to_fitting(
                    job_id=job_id,
                    time_step_start=start,
                    time_step_end=end,
                    time_step_delta=delta,
                )
        if self.status.finished:
            with self._hdf5.open("output") as hdf_output:
                if "potential" in hdf_output.list_groups():
                    self._potential.from_hdf(hdf=hdf_output)

    def get_suggested_number_of_configuration(
        self, species_count=None, multiplication_factor=2.0
    ):
        if self.input["filepath"] == "auto":
            potential = self.input["potential"]
            if isinstance(potential, numbers.Integral):
                potential = f"{potential:02}g.mtp"
            source = self._find_potential(potential)
        else:
            source = self.input["filepath"]
        with open(source) as f:
            lines = f.readlines()
        radial_basis_size, radial_funcs_count, alpha_scalar_moments = 0.0, 0.0, 0.0
        for line in lines:
            if "radial_basis_size" in line:
                radial_basis_size = int(line.split()[2])
            if "radial_funcs_count" in line:
                radial_funcs_count = int(line.split()[2])
            if "species_count" in line and species_count is None:
                species_count = int(line.split()[2])
            if "alpha_scalar_moments" in line:
                alpha_scalar_moments = int(line.split()[2])
        return int(
            multiplication_factor
            * (
                radial_basis_size * radial_funcs_count * species_count**2
                + alpha_scalar_moments
            )
        )

    @staticmethod
    def _update_species_count(lines, species_count):
        ind = 0
        for ind, line in enumerate(lines):
            if "species_count" in line:
                break
        species_line = lines[ind].split()
        species_line[-1] = str(species_count)
        lines[ind] = " ".join(species_line) + "\n"
        return lines

    @staticmethod
    def _update_min_max_distance(lines, min_distance, max_distance):
        min_dis_updated, max_dis_updated = False, False
        for ind, line in enumerate(lines):
            if "min_dist" in line and min_distance is not None:
                min_dist_line = lines[ind].split()
                min_dist_line[2] = str(min_distance)
                lines[ind] = " ".join(min_dist_line) + "\n"
                min_dis_updated = True
            elif "max_dist" in line and max_distance is not None:
                max_dist_line = lines[ind].split()
                max_dist_line[2] = str(max_distance)
                lines[ind] = " ".join(max_dist_line) + "\n"
                max_dis_updated = True
            elif min_dis_updated and max_dis_updated:
                break
        return lines

    def _copy_potential(self, species_count, file_name, cwd=None):
        if cwd is not None:
            file_name = posixpath.join(cwd, file_name)
        if self.input["filepath"] == "auto":
            potential = self.input["potential"]
            if isinstance(potential, numbers.Integral):
                potential = f"{potential:02}g.mtp"
            source = self._find_potential(potential)
        else:
            source = self.input["filepath"]
        with open(source) as f:
            lines = f.readlines()
        if self.input["filepath"] == "auto":
            lines_modified = self._update_species_count(lines, species_count)
            lines_modified = self._update_min_max_distance(
                lines=lines_modified,
                min_distance=self.input["min_dist"],
                max_distance=self.input["max_dist"],
            )
        else:
            lines_modified = lines
        with open(file_name, "w") as f:
            f.writelines(lines_modified)

    @staticmethod
    def stress_tensor_components(stresses):
        return (
            np.array(
                [
                    stresses[0][0],
                    stresses[1][1],
                    stresses[2][2],
                    stresses[1][2],
                    stresses[0][1],
                    stresses[0][2],
                ]
            )
            * gpa_to_ev_ang
        )

    def _write_configurations(
        self, file_name="training.cfg", cwd=None, respect_step=True
    ):
        if cwd is not None:
            file_name = posixpath.join(cwd, file_name)
        (
            indices_lst,
            position_lst,
            forces_lst,
            cell_lst,
            energy_lst,
            track_lst,
            stress_lst,
        ) = ([], [], [], [], [], [], [])
        all_species = set()
        for job_id, value in self._job_dict.items():
            ham = self.project.inspect(job_id)
            if respect_step:
                start = value["time_step_start"]
                end = value["time_step_end"] + 1
                delta = value["time_step_delta"]
            else:
                start, end, delta = 0, None, 1
            time_step = start
            # HACK: until the training container has a proper HDF5 interface
            if ham.__name__ == "TrainingContainer":
                job = ham.to_object()
                all_species.update(job.get_elements())
                symbol_map = {s: i for i, s in enumerate(sorted(set(all_species)))}
                if end is None:
                    end = job.number_of_structures
                for time_step in range(start, end, delta):
                    symbols = job._container.get_array("symbols", time_step)
                    indices_lst.append([symbol_map[s] for s in symbols])
                    position_lst.append(
                        job._container.get_array("positions", time_step)
                    )
                    forces_lst.append(job._container.get_array("forces", time_step))
                    cell_lst.append(job._container.get_array("cell", time_step))
                    energy_lst.append(job._container.get_array("energy", time_step))
                    if job._container.has_array("stress"):
                        volume = np.abs(np.linalg.det(cell_lst[-1]))
                        stress_lst.append(
                            job._container.get_array("stress", time_step) * volume
                        )
                        if np.isnan(stress_lst[-1]).any():
                            stress_lst[-1] = None
                    track_lst.append(str(ham.job_id) + "_" + str(time_step))
                continue
            original_dict = {
                el: ind for ind, el in enumerate(sorted(ham["input/structure/species"]))
            }
            species_dict = {
                ind: original_dict[el]
                for ind, el in enumerate(ham["input/structure/species"])
            }
            # HACK: this does not preserve element order, but the old code is broken for structures with different
            # species anyway
            all_species.update(ham["input/structure/species"])
            if ham.__name__ in [
                "Vasp",
                "ThermoIntDftEam",
                "ThermoIntDftMtp",
                "ThermoIntVasp",
            ]:
                if len(ham["output/outcar/stresses"]) != 0:
                    for position, forces, cell, energy, stresses, volume in zip(
                        ham["output/generic/positions"][start:end:delta],
                        ham["output/generic/forces"][start:end:delta],
                        ham["output/generic/cells"][start:end:delta],
                        ham["output/generic/dft/energy_free"][start:end:delta],
                        ham["output/outcar/stresses"][start:end:delta],
                        ham["output/generic/volume"][start:end:delta],
                    ):
                        indices_lst.append(
                            [species_dict[el] for el in ham["input/structure/indices"]]
                        )
                        position_lst.append(position)
                        forces_lst.append(forces)
                        cell_lst.append(cell)
                        energy_lst.append(energy)
                        stress_lst.append(stresses * volume / gpa_to_ev_ang)
                        track_lst.append(str(ham.job_id) + "_" + str(time_step))
                        time_step += delta
                else:
                    for position, forces, cell, energy in zip(
                        ham["output/generic/positions"][start:end:delta],
                        ham["output/generic/forces"][start:end:delta],
                        ham["output/generic/cells"][start:end:delta],
                        ham["output/generic/dft/energy_free"][start:end:delta],
                    ):
                        indices_lst.append(
                            [species_dict[el] for el in ham["input/structure/indices"]]
                        )
                        position_lst.append(position)
                        forces_lst.append(forces)
                        cell_lst.append(cell)
                        energy_lst.append(energy)
                        track_lst.append(str(ham.job_id) + "_" + str(time_step))
                        time_step += delta
            elif ham.__name__ in ["Lammps", "LammpsInt2", "LammpsMlip"]:
                for position, forces, cell, energy, stresses, volume in zip(
                    ham["output/generic/positions"][start:end:delta],
                    ham["output/generic/forces"][start:end:delta],
                    ham["output/generic/cells"][start:end:delta],
                    ham["output/generic/energy_pot"][start:end:delta],
                    ham["output/generic/pressures"][start:end:delta],
                    ham["output/generic/volume"][start:end:delta],
                ):
                    indices_lst.append(
                        [species_dict[el] for el in ham["input/structure/indices"]]
                    )
                    position_lst.append(position)
                    forces_lst.append(forces)
                    cell_lst.append(cell)
                    energy_lst.append(energy)
                    stress_lst.append(self.stress_tensor_components(stresses * volume))
                    track_lst.append(str(ham.job_id) + "_" + str(time_step))
                    time_step += delta
            else:
                for position, forces, cell, energy, stresses, volume in zip(
                    ham["output/generic/positions"][start:end:delta],
                    ham["output/generic/forces"][start:end:delta],
                    ham["output/generic/cells"][start:end:delta],
                    ham["output/generic/energy_pot"][start:end:delta],
                    ham["output/generic/pressures"][start:end:delta],
                    ham["output/generic/volume"][start:end:delta],
                ):
                    indices_lst.append(
                        [species_dict[el] for el in ham["input/structure/indices"]]
                    )
                    position_lst.append(position)
                    forces_lst.append(forces)
                    cell_lst.append(cell)
                    energy_lst.append(energy)
                    stress_lst.append(self.stress_tensor_components(stresses * volume))
                    track_lst.append(str(ham.job_id) + "_" + str(time_step))
                    time_step += delta
        write_cfg(
            file_name=file_name,
            indices_lst=indices_lst,
            position_lst=position_lst,
            cell_lst=cell_lst,
            forces_lst=forces_lst,
            energy_lst=energy_lst,
            track_lst=track_lst,
            stress_lst=stress_lst,
        )
        return list(all_species)

    def _write_test_set(self, file_name="testing.cfg", cwd=None):
        return self._write_configurations(
            file_name=file_name, cwd=cwd, respect_step=False
        )

    def _write_training_set(self, file_name="training.cfg", cwd=None):
        self._write_configurations(file_name=file_name, cwd=cwd, respect_step=True)

    @staticmethod
    def _job_dict_lst(job_dict):
        job_id_lst, time_step_start_lst, time_step_end_lst, time_step_delta_lst = (
            [],
            [],
            [],
            [],
        )
        for key, value in job_dict.items():
            job_id_lst.append(key)
            time_step_start_lst.append(value["time_step_start"])
            time_step_end_lst.append(value["time_step_end"])
            time_step_delta_lst.append(value["time_step_delta"])
        return job_id_lst, time_step_start_lst, time_step_end_lst, time_step_delta_lst

    @staticmethod
    def _find_potential(potential_name):
        for resource_path in state.settings.resource_paths:
            if os.path.exists(
                os.path.join(resource_path, "mlip", "potentials", "templates")
            ):
                resource_path = os.path.join(
                    resource_path, "mlip", "potentials", "templates"
                )
            if "potentials" in resource_path and os.path.exists(
                os.path.join(resource_path, potential_name)
            ):
                return os.path.join(resource_path, potential_name)
        raise ValueError("Potential not found!")

    # PotentialFit Implementation
    def _add_training_data(self, container):
        self.add_job_to_fitting(container.id, 0, container.number_of_structures - 1, 1)

    def _get_training_data(self):
        # TODO/BUG: only works after input is written for now, instead this should go over _job_
        return self["input/training_data"].to_object()

    def _get_predicted_data(self):
        return self["output/training_efs"].to_object()

    def get_lammps_potential(self):
        return self.potential_dataframe()


class MlipParameter(GenericParameters):
    def __init__(self, separator_char=" ", comment_char="#", table_name="mlip_inp"):
        super(MlipParameter, self).__init__(
            separator_char=separator_char,
            comment_char=comment_char,
            table_name=table_name,
        )

    def load_default(self, file_content=None):
        if file_content is None:
            file_content = """\
potential 16g.mtp
filepath auto
energy-weight 1
force-weight 0.01
stress-weight 0.001
iteration 1000
min_dist 2.0
max_dist 5.0
"""
        self.load_string(file_content)


class CommandLine(GenericParameters):
    def __init__(self, input_file_name=None, **qwargs):
        super(CommandLine, self).__init__(
            input_file_name=input_file_name,
            table_name="cmd",
            comment_char="#",
            val_only=True,
        )

    def load_default(self, file_content=None):
        if file_content is None:
            file_content = """\
$MLP_COMMAND_PARALLEL train --energy-weight=energy_auto --force-weight=force_auto --stress-weight=stress_auto --max-iter=iteration_auto start.mtp training.cfg > training.log
"""
        self.load_string(file_content)

    def activate_postprocessing(self, file_content=None):
        if file_content is None:
            file_content = """\
$MLP_COMMAND_PARALLEL train --energy-weight=energy_auto --force-weight=force_auto --stress-weight=stress_auto --max-iter=iteration_auto start.mtp training.cfg > training.log
$MLP_COMMAND_SERIAL calc-grade Trained.mtp_ training.cfg testing.cfg grades.cfg --mvs-filename=state.mvs > grading.log
$MLP_COMMAND_SERIAL calc-errors Trained.mtp_ training.cfg > training.errors
$MLP_COMMAND_SERIAL calc-errors Trained.mtp_ testing.cfg > testing.errors
$MLP_COMMAND_SERIAL calc-efs Trained.mtp_ training.cfg training_efs.cfg
$MLP_COMMAND_SERIAL calc-efs Trained.mtp_ testing.cfg testing_efs.cfg
$MLP_COMMAND_SERIAL select-add Trained.mtp_ training.cfg testing.cfg diff.cfg > select.log
cp training.cfg training_new.cfg 
cat diff.cfg >> training_new.cfg
"""
        self.load_string(file_content)


def write_cfg(
    file_name,
    indices_lst,
    position_lst,
    cell_lst,
    forces_lst=None,
    energy_lst=None,
    track_lst=None,
    stress_lst=None,
):
    if stress_lst is None or len(stress_lst) == 0:
        stress_lst = [None] * len(position_lst)
    if forces_lst is None or len(forces_lst) == 0:
        forces_lst = [None] * len(position_lst)
    if track_lst is None or len(track_lst) == 0:
        track_lst = [None] * len(position_lst)
    if energy_lst is None or len(energy_lst) == 0:
        energy_lst = [None] * len(position_lst)
    cfg_lst = []
    for indices, position, forces, cell, energy, stress, track_str in zip(
        indices_lst,
        position_lst,
        forces_lst,
        cell_lst,
        energy_lst,
        stress_lst,
        track_lst,
    ):
        cfg_object = Cfg()
        cfg_object.pos = position
        cfg_object.lat = cell
        cfg_object.types = indices
        cfg_object.energy = energy
        cfg_object.forces = forces
        cfg_object.stresses = stress
        if track_str is not None:
            cfg_object.desc = "pyiron\t" + track_str
        cfg_lst.append(cfg_object)
    savecfgs(filename=file_name, cfgs=cfg_lst, desc=None)


def read_cgfs(file_name):
    cgfs_lst = loadcfgs(filename=file_name, max_cfgs=None)
    cell, positions, forces, stress, energy, indicies, grades, jobids, timesteps = (
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
    )
    for cgf in cgfs_lst:
        if cgf.pos is not None:
            positions.append(cgf.pos)
        if cgf.lat is not None:
            cell.append(cgf.lat)
        if cgf.types is not None:
            indicies.append(cgf.types)
        if cgf.energy is not None:
            energy.append(cgf.energy)
        if cgf.forces is not None:
            forces.append(cgf.forces)
        if cgf.stresses is not None:
            stress.append(
                [
                    [cgf.stresses[0], cgf.stresses[5], cgf.stresses[4]],
                    [cgf.stresses[5], cgf.stresses[1], cgf.stresses[3]],
                    [cgf.stresses[4], cgf.stresses[3], cgf.stresses[2]],
                ]
            )
        if cgf.grade is not None:
            grades.append(cgf.grade)
        if cgf.desc is not None:
            job_id, timestep = cgf.desc.split("_")
            jobids.append(int(job_id))
            timesteps.append(int(timestep))
    return [
        np.array(cell),
        np.array(positions),
        np.array(forces),
        np.array(stress),
        np.array(energy),
        np.array(indicies),
        np.array(grades),
        np.array(jobids),
        np.array(timesteps),
    ]



import abc
from collections import namedtuple
import os.path
from typing import List
from typing import Optional
from sqlalchemy import ForeignKey, String, Integer, Column, create_engine
from sqlalchemy.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.orm import declarative_base, relationship
from sqlalchemy.orm import Session
from sqlalchemy.pool import StaticPool

import pandas as pd

DatabaseEntry = namedtuple(
    "DatabaseEntry", ["name", "username", "project", "status", "jobtype"]
)

Base = declarative_base()


class Project(Base):
    __tablename__ = "project_table"

    id = Column(Integer, primary_key=True)
    location = Column(String(250))
    # too stupid to get the bydirectional thing going, whatevs...
    # jobs = relationship("Job", back_populates="project")
    # jobs = relationship("Job", backref="project")


# FIXME: Can be many-to-many later
class JobStatus(Base):
    __tablename__ = "job_status_table"

    id = Column(Integer, primary_key=True)
    status = Column(String(250))


# FIXME: Can be many-to-many later
class JobType(Base):
    __tablename__ = "job_type_table"

    id = Column(Integer, primary_key=True)
    type = Column(String(250))


class Job(Base):
    __tablename__ = "job_table"

    id = Column(Integer, primary_key=True)
    username = Column(String(250))
    name = Column(String(250))

    jobtype_id = Column(Integer, ForeignKey("job_type_table.id"))
    project_id = Column(Integer, ForeignKey("project_table.id"))
    status_id = Column(Integer, ForeignKey("job_status_table.id"))
    # project = relationship("Project", back_populates="jobs")


# TODO: this will be pyiron_base.IsDatabase
class GenericDatabase(abc.ABC):
    """
    Defines the abstract database interface used by all databases.
    """

    @abc.abstractmethod
    def add_item(self, entry: DatabaseEntry) -> int:
        pass

    @abc.abstractmethod
    def get_item(self, job_id: int) -> DatabaseEntry:
        """
        Return database entry of the specified job.

        Args:
            job_id (int): id of the job

        Returns:
            :class:`.DatabaseEntry`: database entry with the given id

        Raises:
            ValueError: if no job with the given id exists
        """
        pass

    @abc.abstractmethod
    def get_item_id(self, job_name: str, project_id: int) -> Optional[int]:
        pass

    @abc.abstractmethod
    def get_project_id(self, location: str) -> Optional[int]:
        pass

    @abc.abstractmethod
    def update_status(self, job_id, status):
        pass

    @abc.abstractmethod
    def remove_item(self, job_id: int) -> DatabaseEntry:
        pass

    @abc.abstractmethod
    def job_table(self) -> pd.DataFrame:
        pass


class TinyDB(GenericDatabase):
    """
    Minimal database implementation and "reference".  Exists mostly to allow easy testing without messing with
    DatabaseAccess.
    """

    def __init__(self, path, echo=False):
        self._path = path
        self._echo = echo
        kwargs = {}
        if path in (":memory:", ""):
            # this allows to access the same DB from the different threads in one process
            # it's necessary for an in memory database, otherwise all threads see different dbs
            kwargs["poolclass"] = StaticPool
            kwargs["connect_args"] = {"check_same_thread": False}
        self._engine = create_engine(
            f"sqlite:///{self._path}", echo=self._echo, **kwargs
        )
        Base.metadata.create_all(self.engine)
        Base.metadata.reflect(self.engine, extend_existing=True)

    @property
    def engine(self):
        return self._engine

    def add_item(self, entry: DatabaseEntry) -> int:
        with Session(self.engine) as session:
            project = (
                session.query(Project)
                .where(Project.location == entry.project)
                .one_or_none()
            )
            if project is None:
                project = Project(location=entry.project)
                session.add(project)
            jobtype = (
                session.query(JobType)
                .where(JobType.type == entry.jobtype)
                .one_or_none()
            )
            if jobtype is None:
                jobtype = JobType(type=entry.jobtype)
                session.add(jobtype)
            status = JobStatus(status=entry.status)
            session.add(status)
            session.flush()
            job = Job(
                name=entry.name,
                username=entry.username,
                project_id=project.id,
                status_id=status.id,
                jobtype_id=jobtype.id,
            )
            session.add(job)
            session.flush()
            job_id = job.id
            session.commit()
        return job_id

    def update_status(self, job_id, status):
        with Session(self.engine) as session:
            try:
                s = (
                    session.query(JobStatus)
                    .select_from(Job)
                    .where(Job.id == job_id, JobStatus.id == Job.status_id)
                    .one()
                )
                s.status = status
                session.commit()
            except Exception as e:
                raise ValueError(f"job_id {job_id} doesn't exist: {e}")

    def _row_to_entry(self, job_data):
        return DatabaseEntry(
            name=job_data.name,
            project=job_data.location,
            username=job_data.username,
            status=job_data.status,
            jobtype=job_data.type,
        )

    def get_item(self, job_id: int) -> DatabaseEntry:
        """
        Return database entry of the specified job.

        Args:
            job_id (int): id of the job

        Returns:
            :class:`.DatabaseEntry`: database entry with the given id

        Raises:
            ValueError: if no job with the given id exists
        """
        try:
            with Session(self.engine) as session:
                job_data = (
                    session.query(
                        Job.__table__, Project.location, JobStatus.status, JobType.type
                    )
                    .select_from(Job)
                    .where(Job.id == job_id)
                    .join(Project, Job.project_id == Project.id)
                    .join(JobStatus, Job.status_id == JobStatus.id)
                    .join(JobType, Job.jobtype_id == JobType.id)
                    .one()
                )
                return self._row_to_entry(job_data)
        except NoResultFound:
            raise ValueError(f"No job with id {job_id} found!") from None

    def get_item_id(self, job_name: str, project_id: int) -> Optional[int]:
        with Session(self.engine) as session:
            try:
                return (
                    session.query(Job.id)
                    .where(
                        Job.name == job_name,
                        Job.project_id == project_id,
                    )
                    .one()
                    .id
                )
            except (MultipleResultsFound, NoResultFound):
                return None

    def get_project_id(self, location: str) -> Optional[int]:
        with Session(self.engine) as session:
            try:
                return (
                    session.query(Project.id)
                    .where(Project.location == location)
                    .one()
                    .id
                )
            # FIXME: MultipleResultsFound should be reraised because it indicates a broken database
            except (MultipleResultsFound, NoResultFound):
                return None

    def remove_item(self, job_id: int) -> DatabaseEntry:
        # FIXME: probably a bit inefficient, because it makes two connections to the DB
        entry = self.get_item(job_id)
        with Session(self.engine) as session:
            job = session.get(Job, job_id)
            session.delete(job)
            session.commit()
        return entry

    def job_table(self) -> pd.DataFrame:
        with Session(self.engine) as session:
            query = (
                session.query(
                    Job.__table__, Project.location, JobStatus.status, JobType.type
                )
                .select_from(Job)
                .join(Project, Job.project_id == Project.id)
                .join(JobStatus, Job.status_id == JobStatus.id)
                .join(JobType, Job.jobtype_id == JobType.id)
            )
            return pd.DataFrame([r._asdict() for r in query.all()])



from pyiron_contrib.protocol.utils.misc import (
    LoggerMixin,
    ensure_iterable,
    flatten,
    requires_arguments,
    ordered_dict_get_index,
    ordered_dict_get_last,
)
from pyiron_contrib.protocol.utils.dictionaries import (
    IODictionary,
    InputDictionary,
    TimelineDict,
)
from pyiron_contrib.protocol.utils.pointer import Pointer, Path, Crumb, CrumbType
from pyiron_contrib.protocol.utils.event import Event, EventHandler
from pyiron_contrib.protocol.utils.comparers import Comparer



"""Generic Input Base Clases"""

import abc
from copy import deepcopy

from pyiron_contrib.tinybase.storage import HasHDFAdapaterMixin

from pyiron_base.interfaces.object import HasStorage
from pyiron_atomistics.atomistics.structure.has_structure import HasStructure

from ase import Atoms
import numpy as np
import matplotlib.pyplot as plt


class StorageAttribute:
    """
    Create an attribute that is synced to a storage attribute.

    It must be created on a class that derives from :class:`.HasStorage`.  It essentially behaves like a property that
    writes and reads values to the underlying :attr:`.HasStorage.storage`. DataContainer of the class.  When accessing
    this property before setting it, `None` is returned.

    It's possible to modify the default value and the accepted type of values by using the builder-style :meth:`.type`,
    :meth:`.default` and :meth:`.constructor` methods.

    >>> class MyType(HasStorage):
    ...     a = StorageAttribute()
    ...     b = StorageAttribute().type(int)
    ...     c = StorageAttribute().constructor(list)
    ...     d = StorageAttribute().default(42).type(int)

    >>> t = MyType()
    >>> t.a # return None
    >>> t.b = 3
    >>> t.b
    3
    >>> t.b = 'asdf'
    TypeError("'asdf' is not of type <class 'int'>")
    >>> t.c
    []
    >>> t.d
    42
    """

    def __init__(self):
        self.name = None
        self.value_type = object
        self.default_constructor = None

    def __set_name__(self, obj, name):
        self.name = name

    def __get__(self, obj, objtype=None):
        try:
            return obj.storage[self.name]
        except KeyError:
            if self.default_constructor is not None:
                ret = obj.storage[self.name] = self.default_constructor()
                return ret
            else:
                return None

    def __set__(self, obj, value):
        if isinstance(value, self.value_type):
            obj.storage[self.name] = value
        else:
            raise TypeError(f"{value} is not of type {self.value_type}")

    def type(self, value_type):
        """
        Set a type to type check values set on the attribute.

        Args:
            value_type (type, tuple of type): a class or list of classes that are acceptable for the attribute

        Returns:
            self: the object it is called on
        """
        self.value_type = value_type
        return self

    def default(self, value):
        """
        Set a default value, if the attribute is accessed without being set.

        The given value is deep copied before being set.  If your type does not
        support this or it is inefficient, use :meth:`.constructor`.

        Args:
            value: any value to be used as default

        Returns:
            self: the object it is called on
        """
        self.default_constructor = lambda: deepcopy(value)
        return self

    def constructor(self, default_constructor):
        """
        Set a function to create a default value, if the attribute is accessed without being set.

        Args:
            default_constructor (function, type): Either a class or a function that takes no arguments

        Returns:
            self: the object it is called on
        """
        self.default_constructor = default_constructor
        return self

    def doc(self, text):
        self.__doc__ = text
        return self


class AbstractContainer(HasStorage, HasHDFAdapaterMixin, abc.ABC):
    def take(self, other: "AbstractContainer"):
        # TODO: think hard about variance of types
        if not isinstance(self, type(other)):
            raise TypeError("Must pass a superclass to transfer from!")

        mro_iter = {k: v for c in type(other).__mro__ for k, v in c.__dict__.items()}
        for name, attr in mro_iter.items():
            if isinstance(attr, StorageAttribute):
                a = getattr(other, name)
                if a is not None:
                    setattr(self, name, a)

    def put(self, other: "AbstractContainer"):
        other.take(self)


class AbstractInput(AbstractContainer, abc.ABC):
    def check_ready(self):
        return True


class StructureInput(AbstractInput):
    structure = StorageAttribute().type(Atoms)


class MDInput(AbstractInput):
    steps = StorageAttribute().type(int)
    timestep = StorageAttribute().type(float)
    temperature = StorageAttribute().type(float)
    output_steps = StorageAttribute().type(int)


class MinimizeInput(AbstractInput):
    ionic_force_tolerance = StorageAttribute().type(float)
    max_steps = StorageAttribute().type(int)
    output_steps = StorageAttribute().type(int)


class AbstractOutput(AbstractContainer, abc.ABC):
    pass


class EnergyPotOutput(AbstractOutput):
    energy_pot = StorageAttribute().type(float)


class EnergyKinOutput(AbstractOutput):
    energy_kin = StorageAttribute().type(float)


class ForceOutput(AbstractOutput):
    forces = StorageAttribute().type(np.ndarray)


class MDOutput(HasStructure, EnergyPotOutput):
    pot_energies = StorageAttribute().type(list).constructor(list)
    kin_energies = StorageAttribute().type(list).constructor(list)
    forces = StorageAttribute().type(list).constructor(list)
    structures = StorageAttribute().type(list).constructor(list)

    def plot_energies(self):
        plt.plot(self.pot_energies - np.min(self.pot_energies), label="pot")
        plt.plot(self.kin_energies, label="kin")
        plt.legend()

    def _number_of_structures(self):
        return len(self.structures)

    def _get_structure(self, frame, wrap_atoms=True):
        return self.structures[frame]

    # TODO: how to make sure this is generally conforming?
    @property
    def energy_pot(self):
        return self.pot_energies[-1]



from mpi4py import MPI
from fitsnap3lib.fitsnap import FitSnap
from pyiron_base import PythonTemplateJob, DataContainer
from pyiron_lanl.fitsnap.common import ase_scraper


default_input = settings = {
    "BISPECTRUM": {
        "numTypes": 1,
        "twojmax": 8,
        "rcutfac": 4.812302818,
        "rfac0": 0.99363,
        "rmin0": 0.0,
        "wj": 1.0,
        "radelem": 0.5,
        "type": "Be",
        "wselfallflag": 0,
        "chemflag": 0,
        "bzeroflag": 0,
        "quadraticflag": 0,
    },
    "CALCULATOR": {
        "calculator": "LAMMPSSNAP",
        "energy": 1,  # Calculate energy descriptors
        "force": 1,  # Calculate force descriptors
        "stress": 0,  # Calculate virial descriptors
    },
    "REFERENCE": {
        "units": "metal",
        "atom_style": "atomic",
        "pair_style": "hybrid/overlay zero 10.0 zbl 4.0 4.8",
        "pair_coeff1": "* * zero",
        "pair_coeff2": "1 1 zbl 74 74",
    },
    "SOLVER": {"solver": "SVD", "compute_testerrs": 1, "detailed_errors": 1},
    "EXTRAS": {
        "dump_descriptors": 0,
        "dump_truth": 0,
        "dump_weights": 0,
        "dump_dataframe": 0,
    },
    "MEMORY": {"override": 0},
}


class FitsnapJob(PythonTemplateJob):
    def __init__(self, project, job_name):
        super(FitsnapJob, self).__init__(project, job_name)
        self.__version__ = "0.1"
        self.__name__ = "FitsnapJob"
        self.input.update(default_input)
        self._lst_of_struct = []
        self._lst_of_energies = []
        self._lst_of_forces = []
        self._coefficients = []

    @property
    def list_of_structures(self):
        return self._lst_of_struct

    @list_of_structures.setter
    def list_of_structures(self, structure_lst):
        self._lst_of_struct = structure_lst

    @property
    def list_of_energies(self):
        return self._lst_of_energies

    @list_of_energies.setter
    def list_of_energies(self, energies):
        self._lst_of_energies = energies

    @property
    def coefficients(self):
        return self._coefficients

    @property
    def list_of_forces(self):
        return self._lst_of_forces

    @list_of_forces.setter
    def list_of_forces(self, forces):
        self._lst_of_forces = forces

    def run_static(self):
        comm = MPI.COMM_WORLD
        input_dict = self.input.to_builtin()
        snap = FitSnap(input_dict, comm=comm, arglist=["--overwrite"])
        ase_scraper(
            snap, self._lst_of_struct, self._lst_of_energies, self._lst_of_forces
        )
        snap.process_configs()
        snap.solver.perform_fit()
        self._coefficients = snap.solver.fit
        self.status.finished = True

    def to_hdf(self, hdf=None, group_name=None):
        super(FitsnapJob, self).to_hdf(hdf=hdf, group_name=group_name)

    def from_hdf(self, hdf=None, group_name=None):
        super(FitsnapJob, self).from_hdf(hdf=hdf, group_name=group_name)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
from pyiron_contrib.atomistics.mlip.mlip import Mlip, read_cgfs

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class MlipSelect(Mlip):
    def __init__(self, project, job_name):
        super(MlipSelect, self).__init__(project, job_name)
        self.__version__ = None
        self.__name__ = "MlipSelect"
        self._executable_activate()
        del self.input["min_dist"]
        del self.input["max_dist"]
        del self.input["iteration"]
        del self.input["energy-weight"]
        del self.input["force-weight"]
        del self.input["stress-weight"]
        self._command_line._delete_line(0)
        self._command_line._delete_line(0)

    def validate_ready_to_run(self):
        if len(self.restart_file_list) == 0:
            raise ValueError()

    def write_input(self):
        species_count = self._write_test_set(
            file_name="testing.cfg", cwd=self.working_directory
        )
        self._copy_potential(
            species_count, file_name="start.mtp", cwd=self.working_directory
        )
        self._command_line[0] = self._command_line[0].replace(
            "Trained.mtp_", "start.mtp"
        )
        self._command_line.write_file(file_name="mlip.sh", cwd=self.working_directory)

    def collect_output(self):
        file_name = os.path.join(self.working_directory, "diff.cfg")
        (
            cell,
            positions,
            forces,
            stress,
            energy,
            indicies,
            grades,
            jobids,
            timesteps,
        ) = read_cgfs(file_name=file_name)
        with self.project_hdf5.open("output") as hdf5_output:
            hdf5_output["forces"] = forces
            hdf5_output["energy_tot"] = energy
            hdf5_output["pressures"] = stress
            hdf5_output["cells"] = cell
            hdf5_output["positions"] = positions
            hdf5_output["indicies"] = indicies



from pyiron_contrib.tinybase.container import (
    AbstractOutput,
    StructureInput,
    StorageAttribute,
)
from pyiron_contrib.tinybase.task import (
    ListTaskGenerator,
    ListInput,
    ReturnStatus,
)

from copy import deepcopy

import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate as si
import scipy.optimize as so

from pyiron_atomistics.atomistics.structure.has_structure import HasStructure


class MurnaghanInput(StructureInput, ListInput):
    strains = StorageAttribute()
    task = StorageAttribute()

    def check_ready(self):
        structure_ready = self.structure is not None
        strain_ready = len(self.strains) > 0
        task = self.task
        task.input.structure = self.structure
        return structure_ready and strain_ready and task.input.check_ready()

    def set_strain_range(self, volume_range, steps):
        self.strains = (1 + np.linspace(-volume_range, volume_range, steps)) ** (1 / 3)

    def _create_tasks(self):
        cell = self.structure.get_cell()
        tasks = []
        for s in self.strains:
            n = deepcopy(self.task)
            n.input.structure = self.structure.copy()
            n.input.structure.set_cell(cell * s, scale_atoms=True)
            tasks.append(n)
        return tasks


class MurnaghanOutput(AbstractOutput, HasStructure):
    base_structure = StorageAttribute()
    volumes = StorageAttribute().type(np.ndarray)
    energies = StorageAttribute().type(np.ndarray)

    def plot(self, per_atom=True):
        N = len(self.base_structure) if per_atom else 1
        plt.plot(self.volumes / N, self.energies / N)

    @property
    def equilibrium_volume(self):
        inter = si.interp1d(self.volumes, self.energies)
        return so.minimize_scalar(
            inter, bounds=(np.min(self.volumes), np.max(self.volumes))
        ).x

    def _number_of_structures(self):
        return 1

    def _get_structure(self, frame, wrap_atoms=True):
        s = self.base_structure
        s.set_cell(s.get_cell() * (self.equilibrium_volume / s.get_volume()) ** (1 / 3))
        return s


class MurnaghanTask(ListTaskGenerator):
    def _get_input(self):
        return MurnaghanInput()

    def _get_output(self):
        out = MurnaghanOutput()
        out.energies = np.zeros(len(self.input.strains))
        out.volumes = np.zeros(len(self.input.strains))
        out.base_structure = self.input.structure
        return out

    def _extract_output(self, output, step, task, ret, task_output):
        if ret.is_done():
            output.energies[step] = task_output.energy_pot
            output.volumes[step] = task.input.structure.get_volume()



from __future__ import annotations

import inspect
import warnings
from functools import partialmethod
from typing import get_args, get_type_hints, Optional, TYPE_CHECKING

from pyiron_contrib.workflow.channels import InputData, OutputData, NotData
from pyiron_contrib.workflow.has_channel import HasChannel
from pyiron_contrib.workflow.io import Inputs, Outputs, Signals
from pyiron_contrib.workflow.node import Node
from pyiron_contrib.workflow.output_parser import ParseOutput

if TYPE_CHECKING:
    from pyiron_contrib.workflow.composite import Composite
    from pyiron_contrib.workflow.workflow import Workflow


class Function(Node):
    """
    Function nodes wrap an arbitrary python function.
    Node IO, including type hints, is generated automatically from the provided
    function.
    Input data for the wrapped function can be provided as any valid combination of
    `*arg` and `**kwarg` at both initialization and on calling the node.

    On running, the function node executes this wrapped function with its current input
    and uses the results to populate the node output.

    Function nodes must be instantiated with a callable to deterimine their function,
    and a string to name each returned value of that callable. (If you really want to
    return a tuple, just have multiple return values but only one output label -- there
    is currently no way to mix-and-match, i.e. to have multiple return values at least
    one of which is a tuple.)

    The node label (unless otherwise provided), IO channel names, IO types, and input
    defaults for the node are produced _automatically_ from introspection of the node
    function.
    Explicit output labels can be provided to modify the number of return values (from
    $N$ to 1 in case you _want_ a tuple returned) and to dodge constraints on the
    automatic scraping routine (namely, that there be _at most_ one `return`
    expression).
    (Additional properties like storage priority and ontological type are forthcoming
    as kwarg dictionaries with keys corresponding to the channel labels (i.e. the node
    arguments of the node function, or the output labels provided).)

    Actual function node instances can either be instances of the base node class, in
    which case the callable node function *must* be provided OR they can be instances
    of children of this class.
    Those children may define some or all of the node behaviour at the class level, and
    modify their signature accordingly so this is not available for alteration by the
    user, e.g. the node function and output labels may be hard-wired.

    Although not strictly enforced, it is a best-practice that where possible, function
    nodes should be both functional (always returning the same output given the same
    input) and idempotent (not modifying input data in-place, but creating copies where
    necessary and returning new objects as output).
    Further, functions with multiple return branches that return different types or
    numbers of return values may or may not work smoothly, depending on the details.

    By default, function nodes will attempt to run whenever one or more inputs is
    updated, and will attempt to update on initialization (after setting _all_ initial
    input values).

    Output is updated in the `process_run_result` inside the parent class `finish_run`
    call, such that output data gets pushed after the node stops running but before
    then `ran` signal fires: run, process and push result, ran.

    After a node is instantiated, its input can be updated as `*args` and/or `**kwargs`
    on call.
    This invokes an `update()` call, which can in turn invoke `run()` if
    `run_on_updates` is set to `True`.
    `run()` returns the output of the executed function, or a futures object if the
    node is set to use an executor.
    Calling the node or executing an `update()` returns the same thing as running, if
    the node is run, or `None` if it is not set to run on updates or not ready to run.

    Args:
        node_function (callable): The function determining the behaviour of the node.
        label (str): The node's label. (Defaults to the node function's name.)
        run_on_updates (bool): Whether to run when you are updated and all your
            input is ready. (Default is True).
        update_on_instantiation (bool): Whether to force an update at the end of
            instantiation. (Default is True.)
        channels_requiring_update_after_run (list[str]): All the input channels named
            here will be set to `wait_for_update()` at the end of each node run, such
            that they are not `ready` again until they have had their `.update` method
            called. This can be used to create sets of input data _all_ of which must
            be updated before the node is ready to produce output again. (Default is
            None, which makes the list empty.)
        output_labels (Optional[str | list[str] | tuple[str]]): A name for each return
            value of the node function OR a single label. (Default is None, which
            scrapes output labels automatically from the source code of the wrapped
            function.) This can be useful when returned values are not well named, e.g.
            to make the output channel dot-accessible if it would otherwise have a label
            that requires item-string-based access. Additionally, specifying a _single_
            label for a wrapped function that returns a tuple of values ensures that a
            _single_ output channel (holding the tuple) is created, instead of one
            channel for each return value. The default approach of extracting labels
            from the function source code also requires that the function body contain
            _at most_ one `return` expression, so providing explicit labels can be used
            to circumvent this (at your own risk).
        **kwargs: Any additional keyword arguments whose keyword matches the label of an
            input channel will have their value assigned to that channel.

    Attributes:
        inputs (Inputs): A collection of input data channels.
        outputs (Outputs): A collection of output data channels.
        signals (Signals): A holder for input and output collections of signal channels.
        ready (bool): All input reports ready, not running or failed.
        running (bool): Currently running.
        failed (bool): An exception was thrown when executing the node function.
        connected (bool): Any IO channel has at least one connection.
        fully_connected (bool): Every IO channel has at least one connection.

    Methods:
        update: If `run_on_updates` is true and all your input is ready, will
            run the engine.
        run: Parse and process the input, execute the engine, process the results and
            update the output.
        disconnect: Disconnect all data and signal IO connections.

    Examples:
        At the most basic level, to use nodes all we need to do is provide the
        `Function` class with a function and labels for its output, like so:
        >>> from pyiron_contrib.workflow.function import Function
        >>>
        >>> def mwe(x, y):
        ...     return x+1, y-1
        >>>
        >>> plus_minus_1 = Function(mwe)
        >>>
        >>> print(plus_minus_1.outputs["x+1"])
        <class 'pyiron_contrib.workflow.channels.NotData'>

        There is no output because we haven't given our function any input, it has
        no defaults, and we never ran it! It tried to `update()` on instantiation, but
        the update never got to `run()` because the node could see that some its input
        had never been specified. So outputs have the channel default value of
        `NotData` -- a special non-data class (since `None` is sometimes a meaningful
        value in python).

        We'll run into a hiccup if we try to set only one of the inputs and force the
        run:
        >>> plus_minus_1.inputs.x = 2
        >>> plus_minus_1.run()
        TypeError

        This is because the second input (`y`) still has no input value, so we can't do
        the sum.

        Once we update `y`, all the input is ready and the automatic `update()` call
        will be allowed to proceed to a `run()` call, which succeeds and updates the
        output.
        The final thing we need to do is disable the `failed` status we got from our
        last run call
        >>> plus_minus_1.failed = False
        >>> plus_minus_1.inputs.y = 3
        >>> plus_minus_1.outputs.to_value_dict()
        {'x+1': 3, 'y-1': 2}

        We can also, optionally, provide initial values for some or all of the input and
        labels for the output:
        >>> plus_minus_1 = Function(mwe, output_labels=("p1", "m1"),  x=1)
        >>> plus_minus_1.inputs.y = 2  # Automatically triggers an update call now
        >>> plus_minus_1.outputs.to_value_dict()
        {'p1': 2, 'm1': 1}

        Input data can be provided to both initialization and on call as ordered args
        or keyword kwargs.
        When running, updating, or calling the node, the output of the wrapped function
        (if it winds up getting run in the conditional cases of updating and calling) is
        returned:
        >>> plus_minus_1(2, y=3)
        (3, 2)

        Finally, we might stop these updates from happening automatically, even when
        all the input data is present and available:
        >>> plus_minus_1 = Function(
        ...     mwe, output_labels=("p1", "m1"),
        ...     x=0, y=0,
        ...     run_on_updates=False, update_on_instantiation=False
        ... )
        >>> plus_minus_1.outputs.p1.value
        <class 'pyiron_contrib.workflow.channels.NotData'>

        With these flags set, the node requires us to manually call a run:
        >>> plus_minus_1.run()
        (-1, 1)

        So function nodes have the most basic level of protection that they won't run
        if they haven't seen any input data.
        However, we could still get them to raise an error by providing the _wrong_
        data:
        >>> plus_minus_1 = Function(mwe, x=1, y="can't add to an int")
        TypeError

        Here everything tries to run automatically, but we get an error from adding the
        integer and string!
        We can make our node even more sensible by adding type
        hints (and, optionally, default values) when defining the function that the node
        wraps.
        The node will automatically figure out defaults and type hints for the IO
        channels from inspection of the wrapped function.

        In this example, note the mixture of old-school (`typing.Union`) and new (`|`)
        type hints as well as nested hinting with a union-type inside the tuple for the
        return hint.
        Our treatment of type hints is **not infinitely robust**, but covers a wide
        variety of common use cases.
        Note that getting "good" (i.e. dot-accessible) output labels can be achieved by
        using good variable names and returning those variables instead of using
        `output_labels`:
        >>> from typing import Union
        >>>
        >>> def hinted_example(
        ...     x: Union[int, float],
        ...     y: int | float = 1
        ... ) -> tuple[int, int | float]:
        ...     p1, m1 = x+1, y-1
        ...     return p1, m1
        >>>
        >>> plus_minus_1 = Function(hinted_example, x="not an int")
        >>> plus_minus_1.outputs.to_value_dict()
        {'p1': <class 'pyiron_contrib.workflow.channels.NotData'>, 'm1': <class
        'pyiron_contrib.workflow.channels.NotData'>}

        Here, even though all the input has data, the node sees that some of it is the
        wrong type and so the automatic updates don't proceed all the way to a run.
        Note that the type hinting doesn't actually prevent us from assigning bad values
        directly to the channel (although it will, by default, prevent connections
        _between_ type-hinted channels with incompatible hints), but it _does_ stop the
        node from running and throwing an error because it sees that the channel (and
        thus node) is not ready
        >>> plus_minus_1.inputs.x.value
        'not an int'

        >>> plus_minus_1.ready, plus_minus_1.inputs.x.ready, plus_minus_1.inputs.y.ready
        (False, False, True)

        In these examples, we've instantiated nodes directly from the base `Function`
        class, and populated their input directly with data.
        In practice, these nodes are meant to be part of complex workflows; that means
        both that you are likely to have particular nodes that get heavily re-used, and
        that you need the nodes to pass data to each other.

        For reusable nodes, we want to create a sub-class of `Function` that fixes some
        of the node behaviour -- usually the `node_function` and `output_labels`.

        This can be done most easily with the `node` decorator, which takes a function
        and returns a node class:
        >>> from pyiron_contrib.workflow.function import function_node
        >>>
        >>> @function_node(output_labels=("p1", "m1"))
        ... def my_mwe_node(
        ...     x: int | float, y: int | float = 1
        ... ) -> tuple[int | float, int | float]:
        ...     return x+1, y-1
        >>>
        >>> node_instance = my_mwe_node(x=0)
        >>> node_instance.outputs.to_value_dict()
        {'p1': 1, 'm1': 0}

        Where we've passed the output labels and class arguments to the decorator,
        and inital values to the newly-created node class (`my_mwe_node`) at
        instantiation.
        Because we provided a good initial value for `x`, we get our result right away.

        Using the decorator is the recommended way to create new node classes, but this
        magic is just equivalent to these two more verbose ways of defining a new class.
        The first is to override the `__init__` method directly:
        >>> from typing import Literal, Optional
        >>>
        >>> class AlphabetModThree(Function):
        ...     def __init__(
        ...         self,
        ...         label: Optional[str] = None,
        ...         run_on_updates: bool = True,
        ...         update_on_instantiation: bool = False,
        ...         **kwargs
        ...     ):
        ...         super().__init__(
        ...             self.alphabet_mod_three,
        ...             label=label,
        ...             run_on_updates=run_on_updates,
        ...             update_on_instantiation=update_on_instantiation,
        ...             **kwargs
        ...         )
        ...
        ...     @staticmethod
        ...     def alphabet_mod_three(i: int) -> Literal["a", "b", "c"]:
        ...         letter = ["a", "b", "c"][i % 3]
        ...         return letter

        Note that we've overridden the default value for `update_on_instantiation`
        above.
        We can also provide different defaults for these flags as kwargs in the
        decorator.

        The second effectively does the same thing, but leverages python's
        `functools.partialmethod` to do so much more succinctly.
        In this example, note that the function is declared _before_ `__init__` is set,
        so that it is available in the correct scope (above, we could place it
        afterwards because we were accessing it through self).
        >>> from functools import partialmethod
        >>>
        >>> class Adder(Function):
        ...     @staticmethod
        ...     def adder(x: int = 0, y: int = 0) -> int:
        ...         sum = x + y
        ...         return sum
        ...
        ...     __init__ = partialmethod(
        ...         Function.__init__,
        ...         adder,
        ...     )

        Finally, let's put it all together by using both of these nodes at once.
        Instead of setting input to a particular data value, we'll set it to
        be another node's output channel, thus forming a connection.
        When we update the upstream node, we'll see the result passed downstream:
        >>> adder = Adder()
        >>> alpha = AlphabetModThree(i=adder.outputs.sum)
        >>>
        >>> adder.inputs.x = 1
        >>> print(alpha.outputs.letter)
        "b"
        >>> adder.inputs.y = 1
        >>> print(alpha.outputs.letter)
        "c"
        >>> adder.inputs.x = 0
        >>> adder.inputs.y = 0
        >>> print(alpha.outputs.letter)
        "a"

        To see more details on how to use many nodes together, look at the
        `Workflow` class.

    Comments:

        If you use the function argument `self` in the first position, the
        whole node object is inserted there:

        >>> def with_self(self, x):
        >>>     ...
        >>>     return x

        For this function, you don't have the freedom to choose `self`, because
        pyiron automatically sets the node object there (which is also the
        reason why you do not see `self` in the list of inputs).
    """

    def __init__(
        self,
        node_function: callable,
        *args,
        label: Optional[str] = None,
        run_on_updates: bool = True,
        update_on_instantiation: bool = True,
        channels_requiring_update_after_run: Optional[list[str]] = None,
        parent: Optional[Composite] = None,
        output_labels: Optional[str | list[str] | tuple[str]] = None,
        **kwargs,
    ):
        super().__init__(
            label=label if label is not None else node_function.__name__,
            parent=parent,
            run_on_updates=run_on_updates,
            # **kwargs,
        )

        self.node_function = node_function

        self._inputs = None
        self._outputs = None
        self._output_labels = self._get_output_labels(output_labels)
        # TODO: Parse output labels from the node function in case output_labels is None

        self.signals = self._build_signal_channels()

        self.channels_requiring_update_after_run = (
            []
            if channels_requiring_update_after_run is None
            else channels_requiring_update_after_run
        )
        self._verify_that_channels_requiring_update_all_exist()

        self._batch_update_input(*args, **kwargs)

        if update_on_instantiation:
            self.update()

    def _get_output_labels(self, output_labels: str | list[str] | tuple[str] | None):
        """
        If output labels are provided, turn convert them to a list if passed as a
        string and return them, else scrape them from the source channel.

        Note: When the user explicitly provides output channels, they are taking
        responsibility that these are correct, e.g. in terms of quantity, order, etc.
        """
        if output_labels is None:
            return self._scrape_output_labels()
        elif isinstance(output_labels, str):
            return [output_labels]
        else:
            return output_labels

    def _scrape_output_labels(self):
        """
        Inspect the source code to scrape out strings representing the returned values.
        _Only_ works for functions with a single `return` expression in their body.

        Will return expressions and function calls just fine, thus best practice is to
        create well-named variables and return those so that the output labels stay
        dot-accessible.
        """
        parsed_outputs = ParseOutput(self.node_function).output
        return [] if parsed_outputs is None else parsed_outputs

    @property
    def _input_args(self):
        return inspect.signature(self.node_function).parameters

    @property
    def inputs(self) -> Inputs:
        if self._inputs is None:
            self._inputs = Inputs(*self._build_input_channels())
        return self._inputs

    @property
    def outputs(self) -> Outputs:
        if self._outputs is None:
            self._outputs = Outputs(*self._build_output_channels(*self._output_labels))
        return self._outputs

    def _build_input_channels(self):
        channels = []
        type_hints = get_type_hints(self.node_function)

        for ii, (label, value) in enumerate(self._input_args.items()):
            is_self = False
            if label == "self":  # `self` is reserved for the node object
                if ii == 0:
                    is_self = True
                else:
                    warnings.warn(
                        "`self` is used as an argument but not in the first"
                        " position, so it is treated as a normal function"
                        " argument. If it is to be treated as the node object,"
                        " use it as a first argument"
                    )
            if label in self._init_keywords:
                # We allow users to parse arbitrary kwargs as channel initialization
                # So don't let them choose bad channel names
                raise ValueError(
                    f"The Input channel name {label} is not valid. Please choose a "
                    f"name _not_ among {self._init_keywords}"
                )

            try:
                type_hint = type_hints[label]
                if is_self:
                    warnings.warn("type hint for self ignored")
            except KeyError:
                type_hint = None

            default = NotData  # The standard default in DataChannel
            if value.default is not inspect.Parameter.empty:
                if is_self:
                    warnings.warn("default value for self ignored")
                else:
                    default = value.default

            if not is_self:
                channels.append(
                    InputData(
                        label=label,
                        node=self,
                        default=default,
                        type_hint=type_hint,
                    )
                )
        return channels

    @property
    def _init_keywords(self):
        return list(inspect.signature(self.__init__).parameters.keys())

    def _build_output_channels(self, *return_labels: str):
        try:
            type_hints = get_type_hints(self.node_function)["return"]
            if len(return_labels) > 1:
                type_hints = get_args(type_hints)
                if not isinstance(type_hints, tuple):
                    raise TypeError(
                        f"With multiple return labels expected to get a tuple of type "
                        f"hints, but got type {type(type_hints)}"
                    )
                if len(type_hints) != len(return_labels):
                    raise ValueError(
                        f"Expected type hints and return labels to have matching "
                        f"lengths, but got {len(type_hints)} hints and "
                        f"{len(return_labels)} labels: {type_hints}, {return_labels}"
                    )
            else:
                # If there's only one hint, wrap it in a tuple so we can zip it with
                # *return_labels and iterate over both at once
                type_hints = (type_hints,)
        except KeyError:
            type_hints = [None] * len(return_labels)

        channels = []
        for label, hint in zip(return_labels, type_hints):
            channels.append(
                OutputData(
                    label=label,
                    node=self,
                    type_hint=hint,
                )
            )

        return channels

    def _verify_that_channels_requiring_update_all_exist(self):
        if not all(
            channel_name in self.inputs.labels
            for channel_name in self.channels_requiring_update_after_run
        ):
            raise ValueError(
                f"On or more channel name among those listed as requiring updates "
                f"after the node runs ({self.channels_requiring_update_after_run}) was "
                f"not found among the input channels ({self.inputs.labels})"
            )

    @property
    def on_run(self):
        return self.node_function

    @property
    def run_args(self) -> dict:
        kwargs = self.inputs.to_value_dict()
        if "self" in self._input_args:
            if self.executor is not None:
                raise NotImplementedError(
                    f"The node {self.label} cannot be run on an executor because it "
                    f"uses the `self` argument and this functionality is not yet "
                    f"implemented"
                )
            kwargs["self"] = self
        return kwargs

    def process_run_result(self, function_output):
        """
        Take the results of the node function, and use them to update the node output.

        By extracting this as a separate method, we allow the node to pass the actual
        execution off to another entity and release the python process to do other
        things. In such a case, this function should be registered as a callback
        so that the node can finishing "running" and push its data forward when that
        execution is finished.
        """
        for channel_name in self.channels_requiring_update_after_run:
            self.inputs[channel_name].wait_for_update()

        if len(self.outputs) == 0:
            return
        elif len(self.outputs) == 1:
            function_output = (function_output,)

        for out, value in zip(self.outputs, function_output):
            out.update(value)

    def _convert_input_args_and_kwargs_to_input_kwargs(self, *args, **kwargs):
        reverse_keys = list(self._input_args.keys())[::-1]
        if len(args) > len(reverse_keys):
            raise ValueError(
                f"Received {len(args)} positional arguments, but the node {self.label}"
                f"only accepts {len(reverse_keys)} inputs."
            )

        positional_keywords = reverse_keys[-len(args) :] if len(args) > 0 else []  # -0:
        if len(set(positional_keywords).intersection(kwargs.keys())) > 0:
            raise ValueError(
                f"Cannot use {set(positional_keywords).intersection(kwargs.keys())} "
                f"as both positional _and_ keyword arguments; args {args}, kwargs {kwargs}, reverse_keys {reverse_keys}, positional_keyworkds {positional_keywords}"
            )

        for arg in args:
            key = positional_keywords.pop()
            kwargs[key] = arg

        return kwargs

    def _batch_update_input(self, *args, **kwargs):
        kwargs = self._convert_input_args_and_kwargs_to_input_kwargs(*args, **kwargs)
        return super()._batch_update_input(**kwargs)

    def __call__(self, *args, **kwargs) -> None:
        kwargs = self._convert_input_args_and_kwargs_to_input_kwargs(*args, **kwargs)
        return super().__call__(**kwargs)

    def to_dict(self):
        return {
            "label": self.label,
            "ready": self.ready,
            "connected": self.connected,
            "fully_connected": self.fully_connected,
            "inputs": self.inputs.to_dict(),
            "outputs": self.outputs.to_dict(),
            "signals": self.signals.to_dict(),
        }


class Slow(Function):
    """
    Like a regular node, but `run_on_updates` and `update_on_instantiation` default to
    `False`.
    This is intended for wrapping function which are potentially expensive to call,
    where you don't want the output recomputed unless `run()` is _explicitly_ called.
    """

    def __init__(
        self,
        node_function: callable,
        *args,
        label: Optional[str] = None,
        run_on_updates=False,
        update_on_instantiation=False,
        parent: Optional[Workflow] = None,
        output_labels: Optional[str | list[str] | tuple[str]] = None,
        **kwargs,
    ):
        super().__init__(
            node_function,
            *args,
            label=label,
            run_on_updates=run_on_updates,
            update_on_instantiation=update_on_instantiation,
            parent=parent,
            output_labels=output_labels,
            **kwargs,
        )


class SingleValue(Function, HasChannel):
    """
    A node that _must_ return only a single value.

    Attribute and item access is modified to finally attempt access on the output value.
    Note that this means any attributes/method available on the output value become
    available directly at the node level (at least those which don't conflict with the
    existing node namespace).
    """

    def __init__(
        self,
        node_function: callable,
        *args,
        label: Optional[str] = None,
        run_on_updates=True,
        update_on_instantiation=True,
        parent: Optional[Workflow] = None,
        output_labels: Optional[str | list[str] | tuple[str]] = None,
        **kwargs,
    ):
        super().__init__(
            node_function,
            *args,
            label=label,
            run_on_updates=run_on_updates,
            update_on_instantiation=update_on_instantiation,
            parent=parent,
            output_labels=output_labels,
            **kwargs,
        )

    def _get_output_labels(self, output_labels: str | list[str] | tuple[str] | None):
        output_labels = super()._get_output_labels(output_labels)
        if len(output_labels) > 1:
            raise ValueError(
                f"{self.__class__.__name__} must only have a single return value, but "
                f"got multiple output labels: {output_labels}"
            )
        return output_labels

    @property
    def single_value(self):
        return self.outputs[self.outputs.labels[0]].value

    @property
    def channel(self) -> OutputData:
        """The channel for the single output"""
        return list(self.outputs.channel_dict.values())[0]

    def __getitem__(self, item):
        return self.single_value.__getitem__(item)

    def __getattr__(self, item):
        return getattr(self.single_value, item)

    def __repr__(self):
        return self.single_value.__repr__()

    def __str__(self):
        return f"{self.label} ({self.__class__.__name__}) output single-value: " + str(
            self.single_value
        )


def function_node(**node_class_kwargs):
    """
    A decorator for dynamically creating node classes from functions.

    Decorates a function.
    Returns a `Function` subclass whose name is the camel-case version of the function
    node, and whose signature is modified to exclude the node function and output labels
    (which are explicitly defined in the process of using the decorator).

    Optionally takes any keyword arguments of `Function`.
    """

    def as_node(node_function: callable):
        return type(
            node_function.__name__.title().replace("_", ""),  # fnc_name to CamelCase
            (Function,),  # Define parentage
            {
                "__init__": partialmethod(
                    Function.__init__,
                    node_function,
                    **node_class_kwargs,
                )
            },
        )

    return as_node


def slow_node(**node_class_kwargs):
    """
    A decorator for dynamically creating slow node classes from functions.

    Unlike normal nodes, slow nodes do update themselves on initialization and do not
    run themselves when they get updated -- i.e. they will not run when their input
    changes, `run()` must be explicitly called.

    Optionally takes any keyword arguments of `Slow`.
    """

    def as_slow_node(node_function: callable):
        return type(
            node_function.__name__.title().replace("_", ""),  # fnc_name to CamelCase
            (Slow,),  # Define parentage
            {
                "__init__": partialmethod(
                    Slow.__init__,
                    node_function,
                    **node_class_kwargs,
                )
            },
        )

    return as_slow_node


def single_value_node(**node_class_kwargs):
    """
    A decorator for dynamically creating fast node classes from functions.

    Unlike normal nodes, fast nodes _must_ have default values set for all their inputs.

    Optionally takes any keyword arguments of `SingleValueNode`.
    """

    def as_single_value_node(node_function: callable):
        return type(
            node_function.__name__.title().replace("_", ""),  # fnc_name to CamelCase
            (SingleValue,),  # Define parentage
            {
                "__init__": partialmethod(
                    SingleValue.__init__,
                    node_function,
                    **node_class_kwargs,
                )
            },
        )

    return as_single_value_node



"""
Creators of various things.

They serve to give easy access to various classes, so that users do not have to
manually import them.

Rough spec:

    1. RootCreator is merely the entry point and delegates to sub creators
    based on name
    2. Sub creators should be type homogenous or related by functionality, i.e.
    one of jobs, one for tasks, one for structures, etc.
    3. Each project gets its own RootCreator, but should keep that instance
    alive so that sub creators may cache certain things
"""

import abc
import importlib
from typing import Union
from functools import wraps
from os import sched_getaffinity

import pyiron_contrib.tinybase.job
from pyiron_contrib.tinybase.project import JobNotFoundError
from pyiron_contrib.tinybase.executor import (
    Executor,
    ProcessExecutor,
    BackgroundExecutor,
    DaskExecutor,
)
from pyiron_atomistics import ase_to_pyiron, Atoms

import ase.build


class Creator(abc.ABC):
    @abc.abstractmethod
    def __init__(self, project, config):
        pass

    # TODO add all methods


def load_class(class_path):
    module, klass = class_path.rsplit(".", maxsplit=1)
    try:
        return getattr(importlib.import_module(module), klass)
    except ImportError as e:
        raise ValueError(
            f"Importing task class '{class_path}' failed: {e.args}!"
        ) from None


class JobCreator(Creator):
    def __init__(self, project, job_dict):
        self._project = project
        self._jobs = job_dict.copy()

    def __dir__(self):
        return tuple(self._jobs.keys())

    def __getattr__(self, task_type):
        if task_type not in self._jobs:
            raise ValueError(f"Unknown task type {task_type}!")
        task = self._jobs[task_type]
        if isinstance(task, str):
            task = self._jobs[task_type] = load_class(task)

        def create(
            name: str,
            delete_existing_job: bool = False,
            delete_aborted_job: bool = False,
        ):
            """
            Create or load a new job.

            Args:
                name (str): name of the job
                delete_existing_job (bool): if a job of this name and type
                    exists, delete it first
                delete_existing_job (bool): if a job if this name and type
                    exists and its status is aborted, delete it first
            """
            try:
                job = self._project.load(name)
                if delete_existing_job or (
                    delete_aborted_job and job.status == "aborted"
                ):
                    job.remove()
                else:
                    if not isinstance(job.task, task):
                        raise ValueError(
                            "Job with given name already exists, but is of different type!"
                        )
                    return job
            except JobNotFoundError:
                pass
            return pyiron_contrib.tinybase.job.TinyJob(task(), self._project, name)

        return create

    def register(self, task: Union["AbstractTask", str], name: str):
        """
        Register a new task.

        Args:
            task (AbstractTask): new task to register

        Returns:
            AbstractTask: the newly registered task
        """
        if name is None:
            name = type(task).__name__
        if name in self._jobs:
            if task != self._jobs[name]:
                raise ValueError(
                    "Refusing to register task: different task of the same name already exists!"
                )
            else:
                return self._jobs[name]
        self._jobs[name] = task
        return task


class TaskCreator(Creator):
    def __init__(self, project, config):
        self._tasks = config.copy()

    def __dir__(self):
        return tuple(self._tasks.keys())

    def __getattr__(self, name):
        task = self._tasks[name]
        if isinstance(task, str):
            task = self._tasks[name] = load_class(task)
        return task

    def register(self, task: Union["AbstractTask", str], name: str):
        """
        Register a new task.

        Args:
            task (AbstractTask): new task to register

        Returns:
            AbstractTask: the newly registered task
        """
        if name is None:
            name = type(task).__name__
        if name in self._jobs:
            if task != self._jobs[name]:
                raise ValueError(
                    "Refusing to register task: different task of the same name already exists!"
                )
            else:
                return self._jobs[name]
        self._jobs[name] = task
        return task


class StructureCreator(Creator):
    def __init__(self, project, config):
        pass

    @wraps(ase.build.bulk)
    def bulk(self, *args, **kwargs):
        return ase_to_pyiron(ase.build.bulk(*args, **kwargs))

    atoms = Atoms


class ExecutorCreator(Creator):
    _DEFAULT_CPUS = min(int(0.5 * len(sched_getaffinity(0))), 8)

    def __init__(self, project, config):
        self._most_recent = None

    def most_recent(self):
        if self._most_recent is None:
            self._most_recent = Executor()
        return self._most_recent

    def _save(func):
        @wraps(func)
        def f(self, *args, **kwargs):
            self._most_recent = func(self, *args, **kwargs)
            return self._most_recent

        return f

    @wraps(ProcessExecutor)
    @_save
    def process(self, max_processes=_DEFAULT_CPUS):
        return ProcessExecutor(max_processes=max_processes)

    @wraps(BackgroundExecutor)
    @_save
    def background(self, max_threads=4):
        return BackgroundExecutor(max_threads=max_threads)

    @wraps(DaskExecutor.from_localcluster)
    @_save
    def dask_local(self, max_workers=_DEFAULT_CPUS, **kwargs):
        return DaskExecutor.from_localcluster(max_workers=max_workers, **kwargs)

    @wraps(DaskExecutor.from_cluster)
    @_save
    def dask_cluster(self, cluster):
        return DaskExecutor.from_cluster(cluster)

    del _save


class RootCreator:
    def __init__(self, project, config):
        self._project = project
        self._subcreators = {}
        for name, (subcreator, subconfig) in config.items():
            self._subcreators[name] = subcreator(project, subconfig)

    def __dir__(self):
        return tuple(self._subcreators.keys())

    def register(self, name, subcreator, force=False):
        """
        Add a new creator.
        """
        if name not in self._subcreators or force:
            self._subcreators[name] = subcreator
        else:
            raise ValueError(f"Already registered a creator under name {name}!")

    def __getattr__(self, name):
        try:
            return self._subcreators[name]
        except KeyError:
            raise AttributeError(f"No creator of name {name} registered!")


TASK_CONFIG = {
    "AseStatic": "pyiron_contrib.tinybase.ase.AseStaticTask",
    "AseMinimize": "pyiron_contrib.tinybase.ase.AseMinimizeTask",
    "AseMD": "pyiron_contrib.tinybase.ase.AseMDTask",
    "Murnaghan": "pyiron_contrib.tinybase.murn.MurnaghanTask",
}


CREATOR_CONFIG = {
    "job": (JobCreator, TASK_CONFIG),
    "task": (TaskCreator, TASK_CONFIG),
    "structure": (StructureCreator, {}),
    "executor": (ExecutorCreator, {}),
}



"""
Collections of channel objects.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from warnings import warn

from pyiron_contrib.workflow.channels import (
    Channel,
    DataChannel,
    InputData,
    OutputData,
    SignalChannel,
    InputSignal,
    OutputSignal,
)
from pyiron_contrib.workflow.has_channel import HasChannel
from pyiron_contrib.workflow.has_to_dict import HasToDict
from pyiron_contrib.workflow.util import DotDict


class IO(HasToDict, ABC):
    """
    IO is a convenience layer for holding and accessing multiple input/output channels.
    It allows key and dot-based access to the underlying channels.
    Channels can also be iterated over, and there are a number of helper functions to
    alter the properties of or check the status of all the channels at once.

    A new channel can be assigned as an attribute of an IO collection, as long as it
    matches the channel's type (e.g. `OutputChannel` for `Outputs`, `InputChannel`
    for `Inputs`, etc...).

    When assigning something to an attribute holding an existing channel, if the
    assigned object is a `Channel`, then an attempt is made to make a `connection`
    between the two channels, otherwise we fall back on a value assignment that must
    be defined in child classes under `_assign_value_to_existing_channel`, i.e.
    >>> some_io.some_existing_channel = 5

    is equivalent to
    >>> some_io._assign_value_to_existing_channel(
    ...     some_io["some_existing_channel"], 5
    ... )

    and
    >>> some_io.some_existing_channel = some_other_channel

    is equivalent to
    >>> some_io.some_existing_channel.connect(some_other_channel)
    """

    def __init__(self, *channels: Channel):
        self.__dict__["channel_dict"] = DotDict(
            {
                channel.label: channel
                for channel in channels
                if isinstance(channel, self._channel_class)
            }
        )

    @property
    @abstractmethod
    def _channel_class(self) -> type(Channel):
        pass

    @abstractmethod
    def _assign_a_non_channel_value(self, channel: Channel, value) -> None:
        """What to do when some non-channel value gets assigned to a channel"""
        pass

    def __getattr__(self, item) -> Channel:
        return self.channel_dict[item]

    def __setattr__(self, key, value):
        if key in self.channel_dict.keys():
            self._assign_value_to_existing_channel(self.channel_dict[key], value)
        elif isinstance(value, self._channel_class):
            if key != value.label:
                warn(
                    f"Assigning a channel with the label {value.label} to the io key "
                    f"{key}"
                )
            self.channel_dict[key] = value
        else:
            raise TypeError(
                f"Can only set Channel object or connect to existing channels, but the "
                f"attribute {key} got assigned {value} of type {type(value)}"
            )

    def _assign_value_to_existing_channel(self, channel: Channel, value) -> None:
        if isinstance(value, HasChannel):
            channel.connect(value.channel)
        else:
            self._assign_a_non_channel_value(channel, value)

    def __getitem__(self, item) -> Channel:
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    @property
    def connected(self):
        return any([c.connected for c in self])

    @property
    def fully_connected(self):
        return all([c.connected for c in self])

    def disconnect(self):
        for c in self:
            c.disconnect_all()

    @property
    def labels(self):
        return list(self.channel_dict.keys())

    def __iter__(self):
        return self.channel_dict.values().__iter__()

    def __len__(self):
        return len(self.channel_dict)

    def __dir__(self):
        return set(super().__dir__() + self.labels)

    def to_dict(self):
        return {
            "label": self.__class__.__name__,
            "connected": self.connected,
            "fully_connected": self.fully_connected,
            "channels": {l: c.to_dict() for l, c in self.channel_dict.items()},
        }


class DataIO(IO, ABC):
    """
    Extends the base IO class with helper methods relevant to data channels.
    """

    def _assign_a_non_channel_value(self, channel: DataChannel, value) -> None:
        channel.update(value)

    def to_value_dict(self):
        return {label: channel.value for label, channel in self.channel_dict.items()}

    @property
    def ready(self):
        return all([c.ready for c in self])

    def to_dict(self):
        d = super().to_dict()
        d["ready"] = self.ready
        return d


class Inputs(DataIO):
    @property
    def _channel_class(self) -> type(InputData):
        return InputData

    def activate_strict_connections(self):
        [c.activate_strict_connections() for c in self]

    def deactivate_strict_connections(self):
        [c.deactivate_strict_connections() for c in self]


class Outputs(DataIO):
    @property
    def _channel_class(self) -> type(OutputData):
        return OutputData


class SignalIO(IO, ABC):
    def _assign_a_non_channel_value(self, channel: SignalChannel, value) -> None:
        raise TypeError(
            f"Tried to assign {value} ({type(value)} to the {channel.label}, which is "
            f"already a {type(channel)}. Only other signal channels may be connected "
            f"in this way."
        )


class InputSignals(SignalIO):
    @property
    def _channel_class(self) -> type(InputSignal):
        return InputSignal


class OutputSignals(SignalIO):
    @property
    def _channel_class(self) -> type(OutputSignal):
        return OutputSignal


class Signals:
    """
    A meta-container for input and output signal IO containers.

    Attributes:
        input (InputSignals): An empty input signals IO container.
        output (OutputSignals): An empty input signals IO container.
    """

    def __init__(self):
        self.input = InputSignals()
        self.output = OutputSignals()

    def disconnect(self):
        self.input.disconnect()
        self.output.disconnect()

    @property
    def connected(self):
        return self.input.connected or self.output.connected

    @property
    def fully_connected(self):
        return self.input.fully_connected and self.output.fully_connected

    def to_dict(self):
        return {
            "input": self.input.to_dict(),
            "output": self.output.to_dict(),
        }



from pyiron_contrib.tinybase.container import (
    AbstractInput,
    StorageAttribute,
    StructureInput,
    MDInput,
    MinimizeInput,
    EnergyPotOutput,
    MDOutput,
)
from pyiron_contrib.tinybase.task import AbstractTask, ReturnStatus

from ase.md.langevin import Langevin
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase import units
from ase.optimize.lbfgs import LBFGS
from ase.optimize.fire import FIRE
from ase.optimize.gpmin.gpmin import GPMin


class AseInput(AbstractInput):
    calculator = StorageAttribute()


class AseStaticInput(AseInput, StructureInput):
    pass


class AseStaticTask(AbstractTask):
    def _get_input(self):
        return AseStaticInput()

    def _get_output(self):
        return EnergyPotOutput()

    def _execute(self, output):
        structure = self.input.structure
        structure.calc = self.input.calculator
        output.energy_pot = structure.get_potential_energy()


class AseMDInput(AseInput, MDInput):
    pass


class AseMDTask(AbstractTask):
    def _get_input(self):
        return AseMDInput()

    def _get_output(self):
        return MDOutput()

    def _execute(self, output):
        structure = self.input.structure.copy()
        structure.calc = self.input.calculator

        MaxwellBoltzmannDistribution(
            structure, temperature_K=self.input.temperature * 2
        )

        dyn = Langevin(
            structure,
            timestep=self.input.timestep * units.fs,
            temperature_K=self.input.temperature,
            friction=1e-3,
            append_trajectory=True,
        )

        def parse():
            output.structures.append(structure.copy())
            output.pot_energies.append(structure.get_potential_energy())
            output.kin_energies.append(structure.get_kinetic_energy())
            output.forces.append(structure.get_forces())

        parse()
        dyn.attach(parse, interval=self.input.steps // self.input.output_steps)
        dyn.run(self.input.steps)


class AseMinimizeInput(AseInput, StructureInput, MinimizeInput):
    algo = StorageAttribute().type(str).default("LBFGS")
    minimizer_kwargs = StorageAttribute().type(dict).constructor(dict)

    def lbfgs(self, damping=None, alpha=None):
        self.algo = "LBFGS"
        if damping is not None:
            self.minimizer_kwargs["damping"] = damping
        if alpha is not None:
            self.minimizer_kwargs["alpha"] = alpha

    def fire(self):
        self.algo = "FIRE"
        self.minimizer_kwargs = {}

    def gpmin(self):
        self.algo = "GPMIN"
        self.minimizer_kwargs = {}

    def get_ase_optimizer(self, structure):
        return {"LBFGS": LBFGS, "FIRE": FIRE, "GPMIN": GPMin}.get(self.algo)(
            structure, **self.minimizer_kwargs
        )


class AseMinimizeTask(AbstractTask):
    def _get_input(self):
        return AseMinimizeInput()

    def _get_output(self):
        return MDOutput()

    def _execute(self, output):
        structure = self.input.structure.copy()
        structure.calc = self.input.calculator

        opt = self.input.get_ase_optimizer(structure)

        def parse():
            output.structures.append(structure.copy())
            output.pot_energies.append(structure.get_potential_energy())
            output.kin_energies.append(structure.get_kinetic_energy())
            output.forces.append(structure.get_forces())

        opt.attach(parse, interval=self.input.output_steps)
        opt.run(fmax=self.input.ionic_force_tolerance, steps=self.input.max_steps)
        parse()

        max_force = abs(output.forces[-1]).max()
        force_tolerance = self.input.ionic_force_tolerance
        if max_force > force_tolerance:
            return ReturnStatus(
                "not_converged",
                f"force in last step ({max_force}) is larger than tolerance ({force_tolerance})!",
            )
